import re
import time
import torch
import argparse
from datasets import load_dataset
from transformers import AutoTokenizer
from tqdm import tqdm

from train.src.models.gpt import GPTLMHeadModel as BasedGPTLMHeadModel

torch.manual_seed(0)
torch.cuda.manual_seed(0)

# Load pretrained models
def get_model(args, model_name="attn"): 
    print("\nLoading pretrained models...")
    if 'attn' == model_name:
        return BasedGPTLMHeadModel.from_pretrained_hf(
            "hazyresearch/attn-360m"
        ).to("cuda").to(torch.bfloat16)
    elif "based" == model_name:
        return BasedGPTLMHeadModel.from_pretrained_hf(
            "hazyresearch/my-awesome-model",
            device="cuda", 
            implementation='tk',           # choices are [fla_parallel, tk]
            silent=True,           
            inference_bs=args.bs,
        ).to(torch.bfloat16)
    else:
        assert 0, print("Unknown model selected.")


# Setup tokenizer
def get_data(task_name="hazyresearch/based-fda"):
    print(f"Loading dataset...")
    tokenizer = AutoTokenizer.from_pretrained("gpt2")
    tokenizer.padding_side = "left"
    tokenizer.pad_token = " "
    tokenizer.pad_token_id = 220
    ds = load_dataset(task_name)
    return tokenizer, ds


# Consruct the model input
def get_inputs(tokenizer, batch, context_length, task_name):
    input_texts = [b.strip() for b in batch['text']]
    targets = batch['value']
    if any(e in task_name for e in ['squad']): 
        questions = batch['question']
    else:
        questions = batch['key']

    has_answer = True
    short_texts = []
    for context, answer, question in zip(input_texts, targets, questions):
        
        context = context.strip(".")
        if context.lower().endswith(question.lower()):
            context = context[:-len(question.lower())]
        elif context.lower().endswith(question.lower() + ":"):
            context = context[:-(len(question.lower()) + 1)]
        question = question[0].upper() + question[1:]

        doc_tokens = tokenizer.batch_encode_plus([context], return_tensors="pt", padding=True, truncation=False)['input_ids'][0]
        
        answer_pos = -1
        if type(answer) == list: answer = answer[0]
        if(answer == "" or len(answer) <= 1): return instances, new_doc_set
        answer_pattern = re.compile(re.escape(answer), re.IGNORECASE)
        if answer_match := answer_pattern.search(context):
            if answer_pos == -1 or answer_pos > answer_match.start():
                answer_pos = answer_match.start()
        if not has_answer: # e.g., summarization datasets
            answer_pos = 0

        # Convert the answer_pos to a token value
        new_text = context[:answer_pos]
        new_text_toks = tokenizer.batch_encode_plus([new_text], return_tensors="pt",)['input_ids'][0]
        answer_tok_pos = len(new_text_toks)
        if answer_pos == -1 and 'tok_pos' in doc and len(new_text_toks) > context_length:
            answer_tok_pos = doc['tok_pos']

        # pick new bounds
        half_length = context_length // 2
        start = max(0, answer_tok_pos - half_length)
        completed_length = answer_tok_pos - start
        remaining_length = context_length - completed_length
        end = min(len(doc_tokens), answer_tok_pos + remaining_length)
        subset_tokens = doc_tokens[start:end]
        short_context = tokenizer.decode(subset_tokens, skip_special_tokens=True)

        # prompt construction
        if any(e in task_name for e in ["fda", "swde"]): 
            question = question + ":"
        short_context = short_context + ". " + question 
        short_texts.append(short_context)

    input_texts = short_texts
    inputs = tokenizer.batch_encode_plus(
        input_texts, 
        return_tensors="pt", 
        padding=True, 
        truncation=True, 
        max_length=2048
    ).input_ids.to("cuda")
    return inputs, questions, targets


def main(args):
    task_name = args.task_name 
    model_name = args.model_name 
    bs = args.bs
    generation_length, context_length = args.generation_length, args.context_length
    model = get_model(args, model_name)
    tokenizer, ds = get_data(task_name)

    # Collect predictions
    results = []
    total_time = []
    preds_list = []
    values_list = []
    for i in tqdm(range(0, len(ds['validation']), bs)):
        if i > 50: break
        batch = ds['validation'][i:i+bs]
        inputs, questions, targets = get_inputs(tokenizer, batch, context_length, task_name)
        limit = inputs.shape[-1] + generation_length
        start = inputs.shape[-1]

        model.eval()
        fn = model.generate
        with torch.no_grad():
            torch.cuda.synchronize()
            start_time = time.time()
            generations = fn(
                input_ids=inputs,
                max_length=limit,
                temperature=0.1,
                top_k=1,
                top_p=1.0,
                cg=True
            )
            torch.cuda.synchronize()
            end_time = time.time()

        preds = generations[:, start:]
        preds = preds.tolist()
        preds = tokenizer.batch_decode(preds)
        seen_inputs = tokenizer.batch_decode(inputs)
        for inp, p, k, v in zip(seen_inputs, preds, questions, targets):
            correct = v.lower() in p.lower()
            results.append(correct)
            preds_list.append(p.split("\n")[0])
            values_list.append(v.split("\n")[0])
        total_time.append(end_time-start_time)


    # Results
    print(f"{model_name=}: score={sum(results)/len(results)}; total_time={sum(total_time)} seconds")
    output_file = f"{model_name}_predictions_and_values.txt"
    with open(output_file, 'w') as f:
        for i, (pred, value) in enumerate(zip(preds_list, values_list)):
            f.write(f"Example {i}:\n- Prediction: {pred}\n- Value: {value}\n")
    print(f"Predictions and values saved to {output_file}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Model and data configuration")

    parser.add_argument('--bs', type=int, default=64, help='Batch size')
    parser.add_argument('--context_length', type=int, default=1500, help='Context length')
    parser.add_argument('--generation_length', type=int, default=48, help='Generation length')
    parser.add_argument(
        '--model_name', type=str, default="based", choices= ["attn", "based"], help='Name of the model'
    )
    parser.add_argument('--task_name', type=str, default="hazyresearch/based-fda", help='Name of the task/dataset')

    args = parser.parse_args()
    main(args)




import torch 

def __eq(str, x,y, tol=1e-5, debug=False): 
    err = torch.abs(x-y).max()
    pass_str = "pass" if err < tol else "fail" 
    print(f"{str} : {pass_str} [err={err:0.5f}]")
    if(debug and (err > tol)):
        print(f"x\n{x}")
        print(f"y\n{y}")
        print(f"diff\n{x-y}")
        
    return err <= tol


def efficiency(flops, time):
    tflops = flops / 1e12
    time_ms = time / 1e6
    return tflops / time_ms






import torch
from flash_attn_interface import flash_attn_func
import thunderkittens as tk
import random
from tqdm import tqdm
import matplotlib.pyplot as plt
import numpy as np

def pytorch_test(Q, K, V, dO, causal):
    q_ = Q.to(torch.float64).requires_grad_()
    k_ = K.to(torch.float64).requires_grad_()
    v_ = V.to(torch.float64).requires_grad_()
    dO_ = dO.to(torch.float64)
    
    # manual pytorch implementation of scaled dot product attention
    QK = torch.matmul(q_, k_.transpose(-2, -1))
    QK /= (q_.size(-1) ** 0.5)
    if causal:
        mask = torch.triu(torch.ones(QK.size(-2), QK.size(-1)), 1).to(torch.bool).to(QK.device)
        QK.masked_fill_(mask, float('-inf'))
    QK = torch.nn.functional.softmax(QK, dim=-1)
    output = torch.matmul(QK, v_)
    
    output.backward(dO_)
    
    q_grad = q_.grad
    k_grad = k_.grad
    v_grad = v_.grad
    
    q_grad = q_grad.to(torch.bfloat16)
    k_grad = k_grad.to(torch.bfloat16)
    v_grad = v_grad.to(torch.bfloat16)
    output = output.to(torch.bfloat16)
    
    return output, q_grad, k_grad, v_grad

def fa2_test(Q, K, V, dO, causal):
    Q.requires_grad = True
    K.requires_grad = True
    V.requires_grad = True
    output = torch.nn.functional.scaled_dot_product_attention(Q, K, V, is_causal=causal)
    output.backward(dO)
    
    return output, Q.grad, K.grad, V.grad

def fa3_test(Q, K, V, dO, causal):
    Q_  = Q.permute(0, 2, 1, 3).clone().detach().contiguous()
    K_  = K.permute(0, 2, 1, 3).clone().detach().contiguous()
    V_  = V.permute(0, 2, 1, 3).clone().detach().contiguous()
    dO_ = dO.permute(0, 2, 1, 3).clone().detach().contiguous()
    
    Q_.requires_grad = True
    K_.requires_grad = True
    V_.requires_grad = True
    
    out, _ = flash_attn_func(Q_, K_, V_, causal=causal)
    out.backward(dO_)
    
    qgrad = Q_.grad
    kgrad = K_.grad
    vgrad = V_.grad
    
    qg_ = qgrad.permute(0, 2, 1, 3).contiguous()
    kg_ = kgrad.permute(0, 2, 1, 3).contiguous()
    vg_ = vgrad.permute(0, 2, 1, 3).contiguous()
    
    output = out.permute(0, 2, 1, 3).contiguous()
    
    return output, qg_, kg_, vg_

def h100_fwd_kernel_test(Q, K, V, dO, causal, mode): 
    if mode == 'backward':
        o = torch.nn.functional.scaled_dot_product_attention(Q, K, V, is_causal=causal)
        
        q_ = Q.to(torch.float64)
        k_ = K.to(torch.float64)
        
        QK = torch.einsum('bhnd,bhmd->bhnm', q_, k_)
        
        if causal:
            mask = torch.triu(torch.ones(QK.size(-2), QK.size(-1)), 1).to(torch.bool).to(QK.device)
            QK.masked_fill_(mask, float('-inf'))
        
        # compute rowmax
        max_vec = QK.max(dim=-1, keepdim=True).values
        
        QK = QK * (1.0 / (q_.size(-1) ** 0.5))
        QK = QK * (1.44269504089)
        
        max_vec = max_vec * (1.44269504089) * (1.0 / (q_.size(-1) ** 0.5))

        QK = QK - max_vec
        QK = torch.exp2(QK)
        
        norm_vec = QK.sum(dim=-1, keepdim=True)
        
        max_vec  = max_vec * 0.69314718056
        norm_vec = torch.log(norm_vec)
        l_vec   = max_vec + norm_vec
        
        if (q_.size(-1) == 64):
            l_vec = l_vec * -8.0
        if (q_.size(-1) == 128):
            l_vec = l_vec * -11.313708499
        l_vec = l_vec.to(torch.float)
        
        _, l_vec_fa3 = flash_attn_func(Q.permute(0, 2, 1, 3), K.permute(0, 2, 1, 3), V.permute(0, 2, 1, 3), causal=causal)
        if (q_.size(-1) == 64):
            l_vec_fa3 = l_vec_fa3 * -8.0
        if (q_.size(-1) == 128):
            l_vec_fa3 = l_vec_fa3 * -11.313708499
        l_vec_fa3 = l_vec_fa3.to(torch.float)
        
        qg, kg, vg = tk.mha_backward(Q, K, V, o, l_vec_fa3, dO, causal)
        
        return o, qg, kg, vg
    else:
        Q.requires_grad = True
        K.requires_grad = True
        V.requires_grad = True
        
        o, l_vec   = tk.mha_forward(Q, K, V, causal)
        qg, kg, vg = tk.mha_backward(Q, K, V, o, l_vec, dO, causal)
        
        return o, qg, kg, vg

def generate_tensor(shape, mean, std, dtype, device):
    tensor = torch.randn(shape, dtype=dtype, device=device)

    magnitude = torch.norm(tensor, dim=-1, keepdim=True)
    scaled_tensor = tensor * (torch.randn(magnitude.shape, dtype=dtype, device=device) * std + mean) / magnitude
    
    return scaled_tensor.contiguous()

def check_correctness(b, h, n, d, causal, mean, std, num_iterations=100, error_mode='all'):
    results = {
        'TK vs PT': {'sum_diff': 0, 'sum_abs': 0, 'max_diff': 0},
        'FA2 vs PT': {'sum_diff': 0, 'sum_abs': 0, 'max_diff': 0},
        'FA3 vs PT': {'sum_diff': 0, 'sum_abs': 0, 'max_diff': 0}
    }

    for _ in range(num_iterations):
        torch.manual_seed(0)
        
        Q  = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
        K  = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
        V  = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
        dO = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
        
        pt_o, pt_qg, pt_kg, pt_vg = pytorch_test(Q, K, V, dO, causal)
        fa2_o, fa2_qg, fa2_kg, fa2_vg = fa2_test(Q, K, V, dO, causal)
        fa3_o, fa3_qg, fa3_kg, fa3_vg = fa3_test(Q, K, V, dO, causal)
        tk_o, tk_qg, tk_kg, tk_vg = h100_fwd_kernel_test(Q, K, V, dO, causal, error_mode)
        
        if error_mode == 'output':
            tensors = [(pt_o, tk_o, fa2_o, fa3_o)]
        elif error_mode == 'backward':
            tensors = [(pt_qg, tk_qg, fa2_qg, fa3_qg),
                       (pt_kg, tk_kg, fa2_kg, fa3_kg),
                       (pt_vg, tk_vg, fa2_vg, fa3_vg)]
        else:  # 'all'
            tensors = [(pt_o, tk_o, fa2_o, fa3_o),
                       (pt_qg, tk_qg, fa2_qg, fa3_qg),
                       (pt_kg, tk_kg, fa2_kg, fa3_kg),
                       (pt_vg, tk_vg, fa2_vg, fa3_vg)]
        
        for pt, tk, fa2, fa3 in tensors:
            for name, impl in [('TK vs PT', tk), ('FA2 vs PT', fa2), ('FA3 vs PT', fa3)]:
                diff = pt - impl
                abs_diff = torch.abs(diff)
                results[name]['sum_diff'] += torch.sum(abs_diff).item()
                results[name]['sum_abs'] += torch.sum(torch.abs(pt)).item()
                results[name]['max_diff'] = max(results[name]['max_diff'], torch.max(abs_diff).item())
                
        torch.cuda.empty_cache()

    total_elements = b * h * n * d * num_iterations * (1 if error_mode == 'output' else 3 if error_mode == 'backward' else 4)
    for name, data in results.items():
        avg_diff = data['sum_diff'] / total_elements
        max_diff = data['max_diff']
        results[name] = {'avg_diff': avg_diff, 'max_diff': max_diff}

    return results

def generate_error_graphs(b, h, d, causal, mean, std, error_mode='all'):
    seq_lengths = [768 * (2**i) for i in range(3)]

    tk_avg_errors, tk_max_errors = [], []
    fa2_avg_errors, fa2_max_errors = [], []
    fa3_avg_errors, fa3_max_errors = [], []

    for n in tqdm(seq_lengths, desc="Generating error data"):
        results = check_correctness(b, h, n, d, causal, mean, std, error_mode=error_mode)
        
        tk_avg_errors.append(results['TK vs PT']['avg_diff'])
        tk_max_errors.append(results['TK vs PT']['max_diff'])
        fa2_avg_errors.append(results['FA2 vs PT']['avg_diff'])
        fa2_max_errors.append(results['FA2 vs PT']['max_diff'])
        fa3_avg_errors.append(results['FA3 vs PT']['avg_diff'])
        fa3_max_errors.append(results['FA3 vs PT']['max_diff'])

    # Generate average error graph
    plt.figure(figsize=(12, 8))
    plt.plot(seq_lengths, tk_avg_errors, label='TK',   marker='o')
    plt.plot(seq_lengths, fa2_avg_errors, label='FA2', marker='s')
    plt.plot(seq_lengths, fa3_avg_errors, label='FA3', marker='^')

    plt.xlabel('Sequence Length')
    plt.ylabel('Average Error')
    plt.title(f'Average Error vs Sequence Length (b={b}, h={h}, d={d}, causal={causal}, mean={mean}, std={std}, mode={error_mode})')
    plt.legend()
    plt.xscale('log', base=2)
    plt.yscale('linear')
    plt.xticks(seq_lengths, labels=seq_lengths)
    plt.grid(True)
    plt.savefig(f'avg_error_graph_b{b}_h{h}_d{d}_causal{causal}_mean{mean}_std{std}_mode{error_mode}.png')
    plt.close()

    # Generate max error graph
    plt.figure(figsize=(12, 8))
    plt.plot(seq_lengths, tk_max_errors,  label='TK',  marker='o')
    plt.plot(seq_lengths, fa2_max_errors, label='FA2', marker='s')
    plt.plot(seq_lengths, fa3_max_errors, label='FA3', marker='^')

    plt.xlabel('Sequence Length')
    plt.ylabel('Maximum Error')
    plt.title(f'Maximum Error vs Sequence Length (b={b}, h={h}, d={d}, causal={causal}, mean={mean}, std={std}, mode={error_mode})')
    plt.legend()
    plt.xscale('log', base=2)
    plt.yscale('linear')
    plt.xticks(seq_lengths, labels=seq_lengths)
    plt.grid(True)
    plt.savefig(f'max_error_graph_b{b}_h{h}_d{d}_causal{causal}_mean{mean}_std{std}_mode{error_mode}.png')
    plt.close()

# Example usage
b, h, d = 12, 12, 64
causal = True
mean = 1e-1
std = 10

for mode in ['output', 'backward', 'all']:
    generate_error_graphs(b, h, d, causal, mean, std, error_mode=mode)

print("Error graphs generated and saved for all modes.")


import torch
tile = 16
#####
# Helpers
#   
def __eq(str, x,y, tol=1e-5, debug=False): 
    err = torch.abs(x-y).max()
    pass_str = "pass" if err < tol else "fail" 
    print(f"{str} : {pass_str} [err={err:0.5f}]")
    if(debug and (err > tol)):
        print(f"x\n{x}")
        print(f"y\n{y}")
        print(f"diff\n{x-y}")
        
    return err <= tol

def _rtile(b,n,d,dt): return torch.randn(b,n,d,device='cuda', dtype=dt)/(n*d)
def _rhtile(b,h,n,d,dt): return torch.randn(b,h,n,d,device='cuda', dtype=dt)/(n*d)
def _rones(b,n,d,dt): return torch.ones(b,n,d,device='cuda', dtype=dt)

def print_tiles(str, t):
    for i in range(t.size(0)):
        for j in range(t.size(1)//tile):
            print(f"{str} TILE batch={i} tile={j}")
            print(f"{t[i,j*tile:(j+1)*tile,:]}")


import torch
import numpy as np
import thunderkittens as tk


def ref_fftconv(u, k, N):
    L = u.shape[-1]
    u_f = torch.fft.fft(u.float(), n = N)
    k_f = torch.fft.fft(k.float(), n = N)
    y_f = u_f * k_f
    y = torch.fft.ifft(y_f, n = N).real[..., :L].to(u.dtype).contiguous()
    return y

def fft_matrix(N):
    n = torch.arange(N)
    k = n.view(-1, 1)
    M = torch.exp(-2j * torch.pi * n * k / N)
    return M

def compute_twiddle_factors_fft(n, m):
    """Compute the twiddle factors of size n x m"""
    n_a = torch.arange(n).view(-1, 1)
    m_a = torch.arange(m)
    N = n * m
    M = torch.exp(-2j * torch.pi * n_a * m_a / N)
    return M

def ifft_matrix(N):
    n = torch.arange(N)
    k = n.view(-1, 1)
    M = torch.exp(2j * torch.pi * n * k / N)
    return M

def compute_twiddle_factors_ifft(n, m):
    """Compute the twiddle factors of size n x m"""
    n_a = torch.arange(n).view(-1, 1)
    m_a = torch.arange(m)
    N = n * m
    M = torch.exp(2j * torch.pi * n_a * m_a / N)
    return M

def pytorch_test_1024(u, k, TESTNAME='all'):
    u_reals, u_imags = [], []
    u_real = u.to(torch.bfloat16)
    u_imag = torch.zeros_like(u, dtype=torch.bfloat16)
    u_reals.append(u_real)
    u_imags.append(u_imag)
    for i in range(3):
        u_new = torch.randn_like(u)*(i+2)
        u_real = u_new.to(torch.bfloat16)
        u_imag = torch.zeros_like(u, dtype=torch.bfloat16)
        u_reals.append(u_real)
        u_imags.append(u_imag)

    # fft matrix
    f_mat = fft_matrix(N1)
    f_real = f_mat.real.to(torch.bfloat16).contiguous()
    f_imag = f_mat.imag.to(torch.bfloat16).contiguous()

    f_sqrt_64x64_real = torch.zeros(64, 64, device=f_real.device).to(f_real.dtype)
    f_sqrt_64x64_real[:N1, :N1] = f_real
    f_sqrt_64x64_real[N1:, N1:] = f_real

    f_sqrt_64x64_imag = torch.zeros(64, 64, device=f_imag.device).to(f_imag.dtype)
    f_sqrt_64x64_imag[N1:, N1:] = f_imag
    f_sqrt_64x64_imag[:N1, :N1] = f_imag

    # ifft matrix
    finv_mat = ifft_matrix(N1)
    finv_real = finv_mat.real.to(torch.bfloat16).contiguous()
    finv_imag = finv_mat.imag.to(torch.bfloat16).contiguous()

    f_inv_sqrt_64x64_real = torch.zeros(64, 64, device=finv_real.device).to(finv_real.dtype)
    f_inv_sqrt_64x64_real[:N1, :N1] = finv_real
    f_inv_sqrt_64x64_real[N1:, N1:] = finv_real

    f_inv_sqrt_64x64_imag = torch.zeros(64, 64, device=finv_imag.device).to(finv_imag.dtype)
    f_inv_sqrt_64x64_imag[N1:, N1:] = finv_imag
    f_inv_sqrt_64x64_imag[:N1, :N1] = finv_imag

    # Normalization factor to make IFFT exact inverse of FFT
    tw = compute_twiddle_factors_fft(N1, N1) / N
    tw_real = tw.real.to(torch.bfloat16).contiguous()
    tw_imag = tw.imag.to(torch.bfloat16).contiguous()

    tw_real_64x64_real = torch.zeros(64, 64, device=tw_real.device).to(tw_real.dtype)
    tw_real_64x64_real[:N1, :N1] = tw_real
    tw_real_64x64_real[N1:, N1:] = tw_real
    tw_real_64x64_real[:N1, N1:] = tw_real
    tw_real_64x64_real[N1:, :N1] = tw_real

    tw_imag_64x64_imag = torch.zeros(64, 64, device=tw_imag.device).to(tw_imag.dtype)
    tw_imag_64x64_imag[:N1, :N1] = tw_imag
    tw_imag_64x64_imag[N1:, N1:] = tw_imag
    tw_imag_64x64_imag[:N1, N1:] = tw_imag
    tw_imag_64x64_imag[N1:, :N1] = tw_imag

    # twiddle inverse
    twinv = compute_twiddle_factors_ifft(N1, N1)
    twinv_real = twinv.real.to(torch.bfloat16).contiguous()
    twinv_imag = twinv.imag.to(torch.bfloat16).contiguous()

    twinv_real_64x64_real = torch.zeros(64, 64, device=twinv_real.device).to(twinv_real.dtype)
    twinv_real_64x64_real[:N1, :N1] = twinv_real
    twinv_real_64x64_real[N1:, N1:] = twinv_real
    twinv_real_64x64_real[:N1, N1:] = twinv_real
    twinv_real_64x64_real[N1:, :N1] = twinv_real

    twinv_imag_64x64_imag = torch.zeros(64, 64, device=twinv_imag.device).to(twinv_imag.dtype)
    twinv_imag_64x64_imag[:N1, :N1] = twinv_imag
    twinv_imag_64x64_imag[N1:, N1:] = twinv_imag
    twinv_imag_64x64_imag[:N1, N1:] = twinv_imag
    twinv_imag_64x64_imag[N1:, :N1] = twinv_imag

    # Compute the regular FFT if the seq len isn't 512 or 2048
    k_f = torch.fft.fft(k.float(), n = N)
    k_fT = k_f.reshape(H, N1, N1).transpose(-1, -2)
    kfT_real = k_fT.real.to(torch.bfloat16).contiguous()
    kfT_imag = k_fT.imag.to(torch.bfloat16).contiguous()

    k_fT_64x64_real = torch.zeros(H, 64, 64, device=kfT_real.device).to(torch.bfloat16)
    k_fT_64x64_real[:, :N1, :N1] = kfT_real.to(k_fT_64x64_real.dtype)
    k_fT_64x64_real[:, N1:, N1:] = kfT_real.to(k_fT_64x64_real.dtype)
    k_fT_64x64_real[:, :N1, N1:] = kfT_real.to(k_fT_64x64_real.dtype)
    k_fT_64x64_real[:, N1:, :N1] = kfT_real.to(k_fT_64x64_real.dtype)

    k_fT_64x64_imag = torch.zeros(H, 64, 64, device=kfT_imag.device).to(torch.bfloat16)
    k_fT_64x64_imag[:, :N1, :N1] = kfT_imag.to(k_fT_64x64_imag.dtype)
    k_fT_64x64_imag[:, N1:, N1:] = kfT_imag.to(k_fT_64x64_imag.dtype)
    k_fT_64x64_imag[:, :N1, N1:] = kfT_imag.to(k_fT_64x64_imag.dtype)
    k_fT_64x64_imag[:, N1:, :N1] = kfT_imag.to(k_fT_64x64_imag.dtype)

    o_reals = []
    for i in range(4):
        o_real = ref_fftconv(u_reals[i], k, N)
        o_real = o_real.reshape(B, H, N1, N1).to(torch.bfloat16).contiguous()
        o_reals.append(o_real)

    u_reals_reshaped, u_imags_reshaped = [], []
    for i in range(4):
        u_reals_reshaped.append(u_reals[i].reshape(B, H, N1, N1).to(torch.bfloat16).contiguous())
        u_imags_reshaped.append(u_imags[i].reshape(B, H, N1, N1).to(torch.bfloat16).contiguous())
    
    u_real = torch.cat(u_reals_reshaped, dim=0)
    u_imag = torch.cat(u_imags_reshaped, dim=0)

    return u_real, u_imag, k_fT_64x64_real, k_fT_64x64_imag, f_sqrt_64x64_real, f_sqrt_64x64_imag, f_inv_sqrt_64x64_real, f_inv_sqrt_64x64_imag, tw_real_64x64_real, tw_imag_64x64_imag, twinv_real_64x64_real, twinv_imag_64x64_imag, o_reals


def pytorch_test_4096(u, k, TESTNAME='all'):
    u_real = u.to(torch.bfloat16)
    u_imag = torch.zeros_like(u, dtype=torch.bfloat16)
    f_mat = fft_matrix(N1)
    f_real = f_mat.real.to(torch.bfloat16).contiguous()
    f_imag = f_mat.imag.to(torch.bfloat16).contiguous()
    finv_mat = ifft_matrix(N1)
    finv_real = finv_mat.real.to(torch.bfloat16).contiguous()
    finv_imag = finv_mat.imag.to(torch.bfloat16).contiguous()
    # Normalization factor to make IFFT exact inverse of FFT
    tw = compute_twiddle_factors_fft(N1, N1) / N
    tw_real = tw.real.to(torch.bfloat16).contiguous()
    tw_imag = tw.imag.to(torch.bfloat16).contiguous()
    twinv = compute_twiddle_factors_ifft(N1, N1)
    twinv_real = twinv.real.to(torch.bfloat16).contiguous()
    twinv_imag = twinv.imag.to(torch.bfloat16).contiguous()

    # Compute the regular FFT if the seq len isn't 512 or 2048
    k_f = torch.fft.fft(k.float(), n = N)
    k_fT = k_f.reshape(H, N1, N1).transpose(-1, -2)
    kfT_real = k_fT.real.to(torch.bfloat16).contiguous()
    kfT_imag = k_fT.imag.to(torch.bfloat16).contiguous()

    o_real = ref_fftconv(u, k, N)
    o_real = o_real.reshape(B, H, N1, N1).to(torch.bfloat16).contiguous()

    u_real = u_real.reshape(B, H, N1, N1).to(torch.bfloat16).contiguous()
    u_imag = u_imag.reshape(B, H, N1, N1).to(torch.bfloat16).contiguous()

    return u_real, u_imag, kfT_real, kfT_imag, f_real, f_imag, finv_real, finv_imag, tw_real, tw_imag, twinv_real, twinv_imag, o_real


# run test 

N = 1024
B = 2
H = 4
N1 = int(np.sqrt(N))

if N == 1024:
    pytorch_test = pytorch_test_1024
elif N == 4096:
    pytorch_test = pytorch_test_4096
else:
    raise ValueError("N must be 1024 or 4096")

u = (torch.randn((B, H, N), dtype=torch.bfloat16, device='cuda')).to(torch.float32) / H
k = (torch.randn((H, N), dtype=torch.bfloat16, device='cuda')).to(torch.float32) / H

u_real, u_imag, kfT_real, kfT_imag, f_real, f_imag, finv_real, finv_imag, tw_real, tw_imag, twinv_real, twinv_imag, o_real = pytorch_test(u, k)

# put all the inputs on the device
u_real = u_real.to('cuda').contiguous()
u_imag = u_imag.to('cuda').contiguous()
kfT_real = kfT_real.to('cuda').contiguous()
kfT_imag = kfT_imag.to('cuda').contiguous()
f_real = f_real.to('cuda').contiguous()
f_imag = f_imag.to('cuda').contiguous()
finv_real = finv_real.to('cuda').contiguous()
finv_imag = finv_imag.to('cuda').contiguous()
tw_real = tw_real.to('cuda').contiguous()
tw_imag = tw_imag.to('cuda').contiguous()
twinv_real = twinv_real.to('cuda').contiguous()
twinv_imag = twinv_imag.to('cuda').contiguous()

if N == 1024:
    B = B * 4
    u = torch.complex(u_real.float(), u_imag.float()).reshape(B, H, N1*N1).to('cuda')

out = tk.fftconv(
    u_real, 
    kfT_real, kfT_imag, 
    f_real, f_imag, 
    finv_real, finv_imag,
    tw_real, tw_imag, 
    twinv_real, twinv_imag,
    B, H, N, N1
)
out = out.reshape(B, H, N)
print(f"{u.shape=}, {k.shape=}, {out.shape=}")
ref = ref_fftconv(u, k, N).real.to('cuda')
print(f"{out.shape=}, {ref.shape=}")

# # out is all zero? 
is_all_zero = torch.allclose(out, torch.zeros_like(out), atol=1e-3)
print(f"{is_all_zero=}")    

batch = torch.randint(0, B, (1,))   
channel = torch.randint(0, H, (1,))
print(out[batch, channel, 760:780])
print(ref[batch, channel, 760:780])

diff = out - ref.to(out.dtype)
max_diff = torch.abs(diff).max()
max_diff_pos = torch.argmax(torch.abs(diff))
print(f"max diff: {max_diff}")
print(f"out[{max_diff_pos}] = {out.flatten()[max_diff_pos]}")
print(f"ref[{max_diff_pos}] = {ref.flatten()[max_diff_pos]}")

# write diff to file
diff = diff.float().cpu().numpy()
with open(f'diff_{N}.txt', 'w') as f:
    for i in range(diff.shape[0]):
        for j in range(diff.shape[1]):
            for k in range(diff.shape[2]):
                f.write(repr(diff[i, j, k]))
                f.write(' ')





import os    
import pandas as pd
import numpy as np
from collections import defaultdict

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat

import warnings
warnings.filterwarnings("ignore", message=".*not a leaf Tensor is being accessed.*")
warnings.filterwarnings("ignore", message=".*no current CUDA context.*")

############## Our Imports #############
from utils import efficiency
import attention.implementations as attention
import hedgehog.implementations as hedgehog
import based_attention.implementations as based
import rotary.implementations as rotary
import mamba2.implementations as mamba2
import fftconv.implementations as fftconv
import layernorm.implementations as layernorm
############## Efficiency Measurements #############

b = 16
h = 16
dv = 64

def measure_efficiency(dt, n, method_name, method, verbose=False, torch_compile=True):
    if verbose:
        print(f"{b=}, {n=}, {h=}, {dv=}")

    if 'c=t' in method_name:
        causal = True
        flops = mod.get_flops(b, n, dv, h, causal=('causal'), mode='bwd' if 'bwd' in method_name else 'fwd')
    elif 'c=f' in method_name:
        causal = False
        flops = mod.get_flops(b, n, dv, h, causal=causal, mode='bwd' if 'bwd' in method_name else 'fwd')
    else:
        flops = mod.get_flops(b, n, dv, h)

    outputs, times = method(dt, b, h, n, dv, verbose=verbose, torch_compile=torch_compile)
    times = times * 1000

    eff = efficiency(flops, times)
    if verbose:
        print(f"Method {method_name} -- Efficiency: {eff:.2f} TFLOPS, Time: {times:.4f} us and FLOPS: {flops:.2f}")
    torch.cuda.empty_cache()
    return eff, times

if __name__ == "__main__":
    print("Benchmarking the kernels...")

    verbose = False
    torch_compile = True

    for mod in [
        attention, 
        based, 
        rotary, 
        hedgehog, 
        layernorm, 
        mamba2, 
        fftconv
    ]:
        implementations_list = []
        implementations_fwd = mod.IMPLEMENTATIONS
        implementations_list.append(implementations_fwd)
        name = mod.NAME
        print("============" * 4, name, "============" * 4)

        try:
            implementations_bwd = mod.IMPLEMENTATIONS_BWD
            implementations_list.append(implementations_bwd)
        except:
            pass

        for implementations in implementations_list:
            method2tflops = {}
            method2timing = {}

            for m, method in implementations.items():
                flops_result, timing_result = {},  {}
                if verbose:
                    print(f"Method: {m}")
                for n in [
                    1024 if 'attn' not in m else 768, 
                    2048 if 'attn' not in m else 1536, 
                    4096 if 'attn' in m else 3072,
                    8192 if 'attn' in m else 6144,
                    16384 if 'attn' in m else 12288
                ]:
                    if "conv" in m and n not in [1024, 4096]:
                        # restrict to sizes we have implemented
                        continue
                    if "mamba2_triton" in m and n not in [1024, 2048, 4096, 8192]:
                        # the kernel results in DEVICE_SIDE_ASSERTS
                        continue
                    if "layernorm" in m and dv*h != 1024:
                        # restrict to sizes we have implemented
                        print('skipping layernorm due to incompatible model dim')
                    if verbose:
                        print(f"Sequence Length: {n}")
                    tflops, timing = measure_efficiency(torch.bfloat16, n, m, method, verbose=verbose, torch_compile=torch_compile)
                    if tflops > 0: 
                        flops_result[n] = tflops
                        timing_result[n] = timing
                method2tflops[m] = flops_result
                method2timing[m] = timing_result

            # print table
            df = pd.DataFrame(method2tflops).replace(np.nan, 'OOM', regex=True)
            print(df)



import torch
torch.random.manual_seed(42)
from tqdm import trange
import sys

D_QK = 128
D_VO = 128

def generate_inputs(testname, B, H, N):
    alphas = torch.rand((H,), dtype=torch.float32, device='cuda')*3
    betas = torch.rand((H,), dtype=torch.float32, device='cuda')
    if testname == 'randn':
        q = (torch.randn((B, H, N, D_QK), dtype=torch.bfloat16, device='cuda'))
        k = (torch.randn((B, H, N, D_QK), dtype=torch.bfloat16, device='cuda'))
        v = (torch.randn((B, H, N, D_VO), dtype=torch.bfloat16, device='cuda')) / 5
        qmap = (torch.randn((H, D_QK, 64), dtype=torch.bfloat16, device='cuda'))
        kmap = (torch.randn((H, D_QK, 64), dtype=torch.bfloat16, device='cuda'))
    elif testname == 'ones':
        q = (torch.ones((B, H, N, D_QK), dtype=torch.bfloat16, device='cuda')) / 5
        k = (torch.ones((B, H, N, D_QK), dtype=torch.bfloat16, device='cuda')) / 5
        v = (torch.ones((B, H, N, D_VO), dtype=torch.bfloat16, device='cuda')) / 5
        qmap = (torch.ones((H, D_QK, 64), dtype=torch.bfloat16, device='cuda')) / 20
        kmap = (torch.ones((H, D_QK, 64), dtype=torch.bfloat16, device='cuda')) / 20
    elif testname == 'qk_test': # this test breaks if sequence length is set <128
        q = torch.eye(D_QK).reshape((1,1,D_QK,D_QK)).repeat(B, H, N//D_QK, 1)*10
        q = q.to(dtype=torch.bfloat16, device='cuda')
        k = torch.eye(D_QK).reshape((1,1,D_QK,D_QK)).repeat(B, H, N//D_QK, 1)*10
        k = k.to(dtype=torch.bfloat16, device='cuda')
        v = torch.eye(D_VO).reshape((1,1,D_VO,D_VO)).repeat(B, H, N//D_VO, 1)*10
        v = v.to(dtype=torch.bfloat16, device='cuda')
        qmap = (torch.ones((H, D_QK, 64), dtype=torch.bfloat16, device='cuda')) / 20
        kmap = (torch.ones((H, D_QK, 64), dtype=torch.bfloat16, device='cuda')) / 20
    elif testname == 'v_or':
        q = (torch.randn((B, H, N, D_QK), dtype=torch.bfloat16, device='cuda')) / 5
        k = (torch.randn((B, H, N, D_QK), dtype=torch.bfloat16, device='cuda')) / 5
        v = (torch.arange(D_VO*2, dtype=torch.bfloat16, device='cuda')/D_VO).reshape((1,1,2,-1)).repeat(B, H, N//2, 1) / 5
        qmap = (torch.ones((H, D_QK, 64), dtype=torch.bfloat16, device='cuda')) / 20
        kmap = (torch.ones((H, D_QK, 64), dtype=torch.bfloat16, device='cuda')) / 20
    elif testname == 'dbg':
        q = (torch.ones((B, H, N, D_QK), dtype=torch.bfloat16, device='cuda')) / 5
        k = (torch.ones((B, H, N, D_QK), dtype=torch.bfloat16, device='cuda')) / 5
        v = (torch.ones((B, H, N, D_VO), dtype=torch.bfloat16, device='cuda')) / 10
        qmap = (torch.eye(64, dtype=torch.bfloat16, device='cuda')).repeat((H,2,1))
        kmap = (torch.eye(64, dtype=torch.bfloat16, device='cuda')).repeat((H,2,1))
    else:
        print('bad testname')
        sys.exit(0)
    return q, k, v, qmap, kmap, alphas, betas

def pytorch_softmax_gt(x, map_mat, label=None):

    x = torch.einsum('bhmd,hdn->bhmn', x, map_mat)

    if label is not None:
        with open(label, 'w') as f:
            for row in x[0,0,0:256]:
                f.write(' '.join(map(lambda k: f"{k:8.4f}", row[:256].tolist())) + '\n')
    
    x_pos = x
    x_neg = -x
    
    x_pos_max = torch.amax(x_pos, dim=-1, keepdim=True)
    x_neg_max = torch.amax(x_neg, dim=-1, keepdim=True)
    
    x_pos = x_pos - x_pos_max
    x_neg = x_neg - x_neg_max
    
    x_pos_num = torch.exp(x_pos)
    x_pos_den = torch.sum(torch.exp(x_pos), dim=-1, keepdim=True)
    
    x_neg_num = torch.exp(x_neg)
    x_neg_den = torch.sum(torch.exp(x_neg), dim=-1, keepdim=True)
    
    x_pos = x_pos_num / x_pos_den
    x_neg = x_neg_num / x_neg_den
    
    x = torch.cat([x_pos, x_neg], dim=-1).clamp(min=1e-6)
    
    return x

generator_mat = torch.block_diag(*[torch.ones((64,64), device='cuda')]*256) # 16384 x 16384 should be big enough for most porpoises
generator_mat += torch.roll(generator_mat, -64, -1) # this adds the terracing
lin_mask = torch.tril(1-generator_mat).reshape((1,1,16384,16384))
exp_mask = torch.tril(generator_mat).reshape((1,1,16384,16384))
exp_mask = 10000*exp_mask - 10000 # we actually want to effectively subtract infinity instead with this mask, since it should happen pre-exp

def pytorch_test(Q, K, V, Qmap, Kmap, alphas, betas):
    
    Qs = pytorch_softmax_gt(Q, Qmap)
    Ks = pytorch_softmax_gt(K, Kmap)
    
    a_lin = torch.einsum('bhmd,bhnd->bhmn', Qs, Ks).to(torch.float32)
    a_exp = torch.einsum('bhmd,bhnd->bhmn', Q, K).to(torch.float32)
    # mask
    a_lin *= lin_mask[:,:,:a_lin.shape[2], :a_lin.shape[3]] * alphas.reshape((1,-1,1,1)) # zero out unwanted entries
    a_exp += exp_mask[:,:,:a_exp.shape[2], :a_exp.shape[3]] # subtract infinity off of unwanted entries
    a_exp -= a_exp.amax(dim=-1, keepdim=True)
    a_exp = torch.exp(a_exp / (128**.5)) * betas.reshape((1,-1,1,1))

    a = a_exp + a_lin
    a = (a / (a.sum(dim=-1, keepdim=True)+1e-6)).to(torch.bfloat16) # normalize
    
    out = torch.einsum('bhmn,bhnd->bhmd', a, V).to(torch.bfloat16)
    
    kv_state = torch.einsum('bhlf,bhld->bhfd', Ks[:,:,:-64,:], V[:,:,:-64,:]).to(torch.float32).detach()
    k_state  = Ks[:,:,:-64,:].to(torch.float32).sum(dim=-2).detach()    
    return out, kv_state, k_state

NAMES = ['randn', 'ones', 'qk_test', 'v_or', 'dbg']
TESTNAME = sys.argv[1]

# get inputs
B, H, N = 2, 2, 1024
assert TESTNAME in NAMES, f"bad testname {TESTNAME}"
q, k ,v, qmap, kmap, alphas, betas = generate_inputs(TESTNAME, B, H, N)
out, kv_state, k_state = pytorch_test(q, k, v, qmap, kmap, alphas, betas)

with open(f'{TESTNAME}.txt', 'w') as f:
    qf = q.to(torch.float32).flatten().cpu().numpy().tolist()
    kf = k.to(torch.float32).flatten().cpu().numpy().tolist()
    vf = v.to(torch.float32).flatten().cpu().numpy().tolist()
    qmf = qmap.to(torch.float32).flatten().cpu().numpy().tolist()
    kmf = kmap.to(torch.float32).flatten().cpu().numpy().tolist()
    af = alphas.to(torch.float32).cpu().numpy().tolist()
    bf = betas.to(torch.float32).cpu().numpy().tolist()
    of = out.to(torch.float32).flatten().cpu().numpy().tolist()
    kvf = kv_state.flatten().cpu().numpy().tolist()
    ksf = k_state.flatten().cpu().numpy().tolist()

    # 1. alpha and beta 
    for i in trange(len(af)):
        f.write(repr(af[i]))
        f.write(' ')

    for i in trange(len(bf)):
        f.write(repr(bf[i]))
        f.write(' ')

    # 2. qmap and kmap
    for i in trange(len(qmf)):
        f.write(repr(qmf[i]))
        f.write(' ')

    for i in trange(len(kmf)):
        f.write(repr(kmf[i]))
        f.write(' ')

    # 3. q, k, v, out
    for i in trange(len(qf)):
        f.write(repr(qf[i]))
        f.write(' ')
    
    for i in trange(len(kf)):
        f.write(repr(kf[i]))
        f.write(' ')

    for i in trange(len(vf)):
        f.write(repr(vf[i]))
        f.write(' ')
    
    for i in trange(len(of)):
        f.write(repr(of[i]))
        f.write(' ')
    
    # k then kv state
    for i in trange(len(ksf)):
        f.write(repr(ksf[i]))
        f.write(' ')

    for i in trange(len(kvf)):
        f.write(repr(kvf[i]))
        f.write(' ')



import torch
from tqdm import trange
import numpy as np
import sys
import math

# pip install "grouped-query-attention-pytorch @ git+ssh://git@github.com/fkodom/grouped-query-attention-pytorch.git"
from grouped_query_attention_pytorch.attention import scaled_dot_product_gqa

# only generate a single batch/head of data, which makes file loading much faster.
# it does mean we'll have to check batch/head behavior separately later, but that should be much easier to debug.
B = 1
N = int(sys.argv[1])
D = int(sys.argv[2])

H_QO = int(sys.argv[3])
H_KV = int(sys.argv[4])

causal = False

torch.random.manual_seed(42)
q = (torch.randn((B, H_QO, N, D), dtype=torch.bfloat16, device='cuda')).requires_grad_()
k = (torch.randn((B, H_KV, N, D), dtype=torch.bfloat16, device='cuda')).requires_grad_()
v = (torch.randn((B, H_KV, N, D), dtype=torch.bfloat16, device='cuda')).requires_grad_()
grad_output = (torch.randn((B, H_QO, N, D), dtype=torch.bfloat16, device='cuda'))

# pad seqlen to multiple of 128
o, _ = scaled_dot_product_gqa(
    q.permute(0, 2, 1, 3).contiguous(),
    k.permute(0, 2, 1, 3).contiguous(),
    v.permute(0, 2, 1, 3).contiguous(),
    is_causal=causal,
    need_weights=False,
)
o = o.permute(0, 2, 1, 3).contiguous()

##########################################
### EXACT GQA COMPUTATION FROM LLAMA 3 ###
def repeat_kv(x: torch.Tensor, n_rep: int) -> torch.Tensor:
    """torch.repeat_interleave(x, dim=2, repeats=n_rep)"""
    bs, slen, n_kv_heads, head_dim = x.shape
    if n_rep == 1:
        return x
    return (
        x[:, :, :, None, :]
        .expand(bs, slen, n_kv_heads, n_rep, head_dim)
        .reshape(bs, slen, n_kv_heads * n_rep, head_dim)
    )


keys_l = repeat_kv(k.permute(0, 2, 1, 3), H_QO // H_KV).permute(0, 2, 1, 3)
values_l = repeat_kv(v.permute(0, 2, 1, 3), H_QO // H_KV).permute(0, 2, 1, 3)

scores = torch.matmul(q, keys_l.transpose(2, 3)) / math.sqrt(D)

mask = torch.full((N, N), float('-inf'), device=q.device, dtype=q.dtype)
mask = torch.triu(mask, diagonal=1)
if mask is not None and causal:
    scores = scores + mask  # (bs, n_local_heads, seqlen, cache_len + seqlen)
scores = torch.nn.functional.softmax(scores, dim=-1).type_as(q)
output = torch.matmul(scores, values_l)  # (bs, n_local_heads, seqlen, head_dim)
### EXACT GQA COMPUTATION FROM LLAMA 3 ###
##########################################

# now do backwards computations
o.backward(grad_output)

q_grad = q.grad
k_grad = k.grad
v_grad = v.grad
    
softmax_scale = 1 / math.sqrt(D)
l_vec = torch.empty((B, H_QO, N, N), dtype=torch.bfloat16, device=q.device)

for i in range(H_QO):
    group_idx = i // (H_QO // H_KV)
    l_vec[:, i] = torch.einsum("bnd,bmd->bnm", q[:, i], k[:, group_idx]) * softmax_scale

mask = torch.triu(torch.ones(N, N), diagonal=1).to('cuda').bool().unsqueeze(0).unsqueeze(0).expand(B, H_QO, -1, -1)
if causal:
    l_vec = l_vec.masked_fill(mask, float('-inf'))

max_vec = l_vec.max(dim=-1, keepdim=True).values
l_vec = l_vec - max_vec
l_vec = torch.exp(l_vec)
l_vec_sum = l_vec.sum(dim=-1, keepdim=True)
l_vec = max_vec + torch.log(l_vec_sum)

d_vec = torch.mul(o.to(torch.bfloat16), grad_output.to(torch.bfloat16))
d_vec = d_vec.to(torch.bfloat16).sum(dim=-1, keepdim=True)


print("--------------------------------------")
print("Q shape: ",      q.shape)
print("K shape: ",      k.shape)
print("V shape: ",      v.shape)
print("O shape: ",      o.shape)
print("Q grad shape: ", q_grad.shape)
print("K grad shape: ", k_grad.shape)
print("V grad shape: ", v_grad.shape)
print("L shape: ",      l_vec.shape)
print("D shape: ",      d_vec.shape)
print("--------------------------------------")

# print out avg magnitude of output tensor
print(f'Average magnitude of OUTPUT tensor: {o.abs().mean()}')
print(f'1/100 magnitude of OUTPUT tensor:   {o.abs().mean()/100}')
print(f'Average magnitude of Q_GRAD tensor: {q_grad.abs().mean()}')
print(f'1/100 magnitude of Q_GRAD tensor:   {q_grad.abs().mean()/100}')
print(f'Average magnitude of K_GRAD tensor: {k_grad.abs().mean()}')
print(f'1/100 magnitude of K_GRAD tensor:   {k_grad.abs().mean()/100}')
print(f'Average magnitude of V_GRAD tensor: {v_grad.abs().mean()}')
print(f'1/100 magnitude of V_GRAD tensor:   {v_grad.abs().mean()/100}')
print(f'Average magnitude of L tensor:      {l_vec.abs().mean()}')
print(f'1/100 magnitude of L tensor:        {l_vec.abs().mean()/100}')
print(f'Average magnitude of D tensor:      {d_vec.abs().mean()}')
print(f'1/100 magnitude of D tensor:        {d_vec.abs().mean()/100}')
print("--------------------------------------")

filename = f'randn_{N}N_{D}D_{H_QO}QO_{H_KV}KV'

if causal:
    filename += '_causal'
if H_QO != H_KV:
    filename += '_gqa'

filename += '.txt'

with open(filename, 'w') as f:
    # inputs
    qf = q.to(torch.float32).flatten().detach().cpu().numpy()
    kf = k.to(torch.float32).flatten().detach().cpu().numpy()
    vf = v.to(torch.float32).flatten().detach().cpu().numpy()
    of = o.to(torch.float32).flatten().detach().cpu().numpy()
    
    og_f = grad_output.to(torch.float32).flatten().detach().cpu().numpy()
    
    # intermediate
    l_vecf = l_vec.to(torch.float32).flatten().detach().cpu().numpy()
    d_vecf = d_vec.to(torch.float32).flatten().detach().cpu().numpy()
    
    qg_f = q_grad.to(torch.float32).flatten().detach().cpu().numpy()
    kg_f = k_grad.to(torch.float32).flatten().detach().cpu().numpy()
    vg_f = v_grad.to(torch.float32).flatten().detach().cpu().numpy()
    
    for i in trange(q.shape[0] * q.shape[1] * q.shape[2] * q.shape[3]):
        f.write(repr(float(qf[i])))
        f.write(' ')
    for i in trange(k.shape[0] * k.shape[1] * k.shape[2] * k.shape[3]):
        f.write(repr(float(kf[i])))
        f.write(' ')
    for i in trange(v.shape[0] * v.shape[1] * v.shape[2] * v.shape[3]):
        f.write(repr(float(vf[i])))
        f.write(' ')
    for i in trange(o.shape[0] * o.shape[1] * o.shape[2] * o.shape[3]):
        f.write(repr(float(of[i])))
        f.write(' ')
    for i in trange(l_vec.shape[0] * l_vec.shape[1] * l_vec.shape[2]):
        f.write(repr(float(l_vecf[i])))
        f.write(' ')
    for i in trange(d_vec.shape[0] * d_vec.shape[1] * d_vec.shape[2]):
        f.write(repr(float(d_vecf[i])))
        f.write(' ')
    for i in trange(grad_output.shape[0] * grad_output.shape[1] * grad_output.shape[2] * grad_output.shape[3]):
        f.write(repr(float(og_f[i])))
        f.write(' ')
    for i in trange(q_grad.shape[0] * q_grad.shape[1] * q_grad.shape[2] * q_grad.shape[3]):
        f.write(repr(float(qg_f[i])))
        f.write(' ')
    for i in trange(k_grad.shape[0] * k_grad.shape[1] * k_grad.shape[2] * k_grad.shape[3]):
        f.write(repr(float(kg_f[i])))
        f.write(' ')
    for i in trange(v_grad.shape[0] * v_grad.shape[1] * v_grad.shape[2] * v_grad.shape[3]):
        f.write(repr(float(vg_f[i])))
        f.write(' ')


import os
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat
from functools import partial

try:
    import thunderkittens as tk
    print(f"Successfully imported thunderkittens")
except:
    print("Could not import thunderkittens")

try:
    from rotary.baselines.rotary import RotaryEmbedding
    from rotary.baselines.rotary import apply_rotary_emb
    print(f"Successfully imported RotaryEmbedding")
except:
    RotaryEmbedding = None
    print("Could not import RotaryEmbedding. Please obtain from FlashAttention repo.")
    

#################### Rotary ########################

def get_flops(batch, seqlen, headdim, nheads):
    f = batch*seqlen*nheads*(headdim//2) # mult with cos 
    f += batch*seqlen*nheads*(headdim//2) # mult with sin
    f += batch*seqlen*nheads*(headdim//2) # add rotated values
    return f * 2 # for q and k 


def get_rotary_inputs(b, h, n, dv, dt):
    d_model = h * dv
    head_dim = dv
    rotary_emb_fraction = 1
    rotary_emb_dim = head_dim * rotary_emb_fraction
    rotary_emb_base = 10000
    rotary_max_seqlen = None
    rotary_emb_scale_base=None
    rotary_emb_interleaved=False

    assert RotaryEmbedding is not None, "RotaryEmbedding not imported"

    rotary_emb = RotaryEmbedding(
        rotary_emb_dim,
        base=rotary_emb_base,
        scale_base=rotary_emb_scale_base,
        interleaved=rotary_emb_interleaved,
        device='cuda',
    )

    h = d_model // head_dim
    qkv = torch.randn((b, n, 3, h, head_dim), dtype=dt, device='cuda')
    qkv_emb_flash = rotary_emb(
        qkv, seqlen_offset=0, max_seqlen=rotary_max_seqlen
    )
    sin = rotary_emb._sin_cached
    cos = rotary_emb._cos_cached
    return qkv, cos, sin, rotary_max_seqlen, rotary_emb

def rotate_half(x, interleaved=False):
    if not interleaved:
        x1, x2 = x.chunk(2, dim=-1)
        return torch.cat((-x2, x1), dim=-1)
    else:
        x1, x2 = x[..., ::2], x[..., 1::2]
        return rearrange(torch.stack((-x2, x1), dim=-1), "... d two -> ... (d two)", two=2)

class PytorchRotary(nn.Module):
    def __init__(self):
        super().__init__()
    
    def forward(self, q, k, cos, sin, o_dt, dt, ro_dim, interleaved):
        # for q
        cos = repeat(cos, "... d -> ... 1 (2 d)" if not interleaved else "... d -> ... 1 (d 2)")
        sin = repeat(sin, "... d -> ... 1 (2 d)" if not interleaved else "... d -> ... 1 (d 2)")
        y1 = torch.cat(
            [q[..., :ro_dim] * cos + rotate_half(q[..., :ro_dim], interleaved) * sin, q[..., ro_dim:]],
            dim=-1,
        )
        
        # for k
        y2 = torch.cat(
            [k[..., :ro_dim] * cos + rotate_half(k[..., :ro_dim], interleaved) * sin, k[..., ro_dim:]],
            dim=-1,
        )
        return y1, y2
        
def rotary_test(dt, b, h, n, dv, causal, is_forwards, method_str, num_iters=10, verbose=True, torch_compile=False, **kwargs):
    # Reference: https://github.com/Dao-AILab/flash-attention/blob/898dd4bbf237b24ed8fd2a3d13ee33bd156bfb23/flash_attn/modules/mha.py#L95
    
    pytorch_method = PytorchRotary()
    if torch_compile and method_str == "torch":
        try:
            pytorch_method = torch.compile(pytorch_method)
        except Exception as e:
            print(f"Could not compile pytorch_method: {e}")
            
    for stage in ['warmup', 'timed']:
        start_events = [torch.cuda.Event(enable_timing=True) for _ in range(num_iters)]
        end_events = [torch.cuda.Event(enable_timing=True) for _ in range(num_iters)]

        for i in range(num_iters):
            qkv, cos, sin, rotary_max_seqlen, rotary_emb = get_rotary_inputs(b, h, n, dv, dt)
        
            try:
                if method_str == "flash_triton":
                    q = qkv[:,:,0]
                    k = qkv[:,:,1]

                    torch.cuda.synchronize()
                    start_events[i].record()
                    y = apply_rotary_emb(
                        q, cos, sin
                    )
                    y = apply_rotary_emb(
                        k, cos, sin
                    )
                    end_events[i].record()
                    torch.cuda.synchronize()

                elif method_str == "torch":
                    o_dt = qkv.dtype
                    ro_dim = cos.shape[-1] * 2
                    assert ro_dim <= qkv.shape[-1]
                    interleaved = False
                    q = qkv[:, :, 0]
                    k = qkv[:, :, 1]

                    torch.cuda.synchronize()
                    start_events[i].record()
                    
                    y1, y2 = pytorch_method(q, k, cos, sin, o_dt, dt, ro_dim, interleaved)
                    y = torch.cat((y1, y2), dim=-1)
                    
                    end_events[i].record()
                    torch.cuda.synchronize()

                elif method_str == "tk":
                    q = qkv[:,:,0].transpose(1,2).clone()   # torch.Size([b, n, 3, h, dv])
                    k = qkv[:,:,1].transpose(1,2).clone()
                    cos = cos.contiguous()
                    sin = sin.contiguous()
                    q = q.contiguous()
                    k = k.contiguous()

                    torch.cuda.synchronize()
                    start_events[i].record()
                    y = tk.fused_rotary(q, cos, sin)
                    y = tk.fused_rotary(k, cos, sin)
                    end_events[i].record()
                    torch.cuda.synchronize()

                else:
                    assert 0, f"Unknown method: {method_str}"

            except Exception as e:
                if verbose:
                    print(f"Error in rotary_test: {e}")
                return None, -1

            torch.cuda.empty_cache()

    tot = sum([s.elapsed_time(e) for s, e in zip(start_events, end_events)])/num_iters
    assert not np.isnan(y.detach().float().cpu()).any(), "NaN values detected in output 'out'"
    assert not np.isinf(y.detach().float().cpu()).any(), "Inf values detected in output 'out'"
    return y, tot


IMPLEMENTATIONS = {
    "flash_triton": partial(rotary_test, causal=True, is_forwards=True, method_str="flash_triton"),
    "torch": partial(rotary_test, causal=True, is_forwards=True, method_str="torch"),
    "tk": partial(rotary_test, causal=True, is_forwards=True, method_str="tk"),
}

NAME = "ROTARY"



import torch
from tqdm import trange
import numpy as np
import sys
from einops import rearrange
import math

# only generate a single batch/head of data, which makes file loading much faster.
# it does mean we'll have to check batch/head behavior separately later, but that should be much easier to debug.
B = 1
H = 1
N = 1024*16
D = 16
DV = 64

TESTNAME = sys.argv[1]

if TESTNAME in ['ones_all', 'ones_t0', 'ones_t1', 'ones_t0t1', 'ones_t2']:
    q = (torch.ones((B, H, N, D), dtype=torch.bfloat16, device='cuda')/D).to(torch.float32)
    k = (torch.ones((B, H, N, D), dtype=torch.bfloat16, device='cuda')/D).to(torch.float32)
    v = (torch.ones((B, H, N, DV), dtype=torch.bfloat16, device='cuda')/DV).to(torch.float32)
elif TESTNAME in ['randn_all', 'randn_t0', 'randn_t1', 'randn_t0t1', 'randn_t2']:
    torch.random.manual_seed(42)
    q = (torch.randn((B, H, N, D), dtype=torch.bfloat16, device='cuda')/float(D)**.5).to(torch.float32)
    k = (torch.randn((B, H, N, D), dtype=torch.bfloat16, device='cuda')/float(D)**.5).to(torch.float32)
    v = (torch.randn((B, H, N, DV), dtype=torch.bfloat16, device='cuda')/DV).to(torch.float32)
else:
    print('Invalid test name')
    sys.exit(0)

def pytorch_test(Q, K, V, add_scale = True, add_norm = False, TESTNAME='all'):
    print("Adding scale:", add_scale)
    B, H, L, D = Q.shape

    def make_causal(X):
        (b,h,n,m) = X.shape
        mask= ~(torch.arange(n).view(1,1,n,1) >= torch.arange(n).view(1,1,1,n)).expand(b,h,n,n)
        X[mask] = 0.
        return X

    O   = torch.einsum("bhnd,bhmd->bhnm", Q.to(torch.float32), K.to(torch.float32))**2
    O2  = make_causal(O)
    T2  = torch.einsum(
        "bhnm,bhmd->bhnd", 
        O2.to(torch.float32), 
        V.to(torch.float32)
    ).to(torch.float32)
    T1a = make_causal(
        torch.einsum("bhnd,bhmd->bhnm", 
        Q.to(torch.float32), K.to(torch.float32))
    )
    T1 = torch.einsum(
        "bhnm,bhme->bhne", 
        T1a.to(torch.float32), V.to(torch.float32)
    ).to(torch.float32)
    T0  = V.to(torch.float32).cumsum(dim=2)

    rd  = math.sqrt(D) if add_scale else 1    
    rrd = math.sqrt(rd) if add_scale else 1   
    r2  = math.sqrt(2) if add_scale else 1

    # KV states
    A2 = torch.einsum("bhnd,bhnf,bhne->bhndef",K.to(torch.float32),V.to(torch.float32),K.to(torch.float32)).cumsum(dim=2) / (r2 * rd) 
    A2 = A2[:, :, -1]
    kv_a2 = rearrange(A2, 'b h e f d -> b h (e f) d')
    A1 = torch.einsum("bhnd,bhne->bhnde",K.to(torch.float32),V.to(torch.float32)).cumsum(dim=2)  / rrd
    kv_a1 = A1[:, :, -1].transpose(2, 3)

    o = 0
    o += T0.to(torch.float32)
    o += T1.to(torch.float32) / (rrd * rrd)
    o += T2.to(torch.float32) / (rd * r2 * rd * r2)
    return o.to(torch.bfloat16), kv_a2, kv_a1

o, kv_a2, kv_a1 = pytorch_test(q, k, v, TESTNAME)

with open(f'{TESTNAME}_{B}x{H}x{N}x{D}x{DV}.txt', 'w') as f:
    qf = q.to(torch.float32).flatten().cpu().numpy().tolist()
    kf = k.to(torch.float32).flatten().cpu().numpy().tolist()
    vf = v.to(torch.float32).flatten().cpu().numpy().tolist()
    of = o.to(torch.float32).flatten().cpu().numpy().tolist()
    kv_a1_cumsumf = kv_a1.to(torch.float32).flatten().cpu().numpy().tolist()
    kv_a2_cumsumf = kv_a2.to(torch.float32).flatten().cpu().numpy().tolist()

    for i in trange(B*H*N*D):
        f.write(repr(qf[i]))
        f.write(' ')
    for i in trange(B*H*N*D):
        f.write(repr(kf[i]))
        f.write(' ')
    for i in trange(B*H*N*DV):
        f.write(repr(vf[i]))
        f.write(' ')
    for i in trange(B*H*N*DV):
        f.write(repr(of[i]))
        f.write(' ')

    # kv states
    for i in trange(B*H*D*DV):
        f.write(repr(kv_a1_cumsumf[i]))
        f.write(' ')
    for i in trange(B*H*D*D*DV):
        f.write(repr(kv_a2_cumsumf[i]))
        f.write(' ')

    


import torch 
import torch.nn.functional as F
from einops import rearrange
import numpy as np
from functools import partial

try:
    from mamba2.baselines.ssd_minimal import ssd_minimal_discrete 
except:
    from baselines.ssd_minimal import ssd_minimal_discrete

try:
    import thunderkittens as tk 
    print(f"Successfully imported thunderkittens")
except ImportError:
    print(f'Failed to import thunderkittens; Please "pip setup.py install".')
    tk = None

try:
    from mamba_ssm.ops.triton.ssd_combined import mamba_chunk_scan_combined
    print(f"Successfully imported mamba_chunk_scan_combined")
except ImportError:
    print(f"Failed to import mamba_chunk_scan_combined; Please 'pip install mamba_ssm'.")
    mamba_chunk_scan_combined = None


################### Mamba 2 ####################


def get_flops(b: int, n: int, dv: int, h: int, causal: bool = False) -> int:
    """Calculate FLOPs for Mamba-2 operation.
    
    Args:
        b: Batch size
        n: Sequence length
        dv: Head dimension
        h: Number of heads
        causal: Whether using causal attention
    
    Returns:
        Total number of FLOPs
    """
    # Constants
    chunk_length = 64
    state_dimension = 64
    ngroups = 1
    
    num_chunks = n // chunk_length
    seqlen = n
    nheads = h
    head_dimension = dv
    
    flops = 0
    
    # Mask and decay computations
    flops += seqlen * nheads  # cumsum
    flops += seqlen * nheads * chunk_length  # segsum
    flops += seqlen * nheads * chunk_length  # exp
    
    # Center blocks
    flops += 2 * num_chunks * ngroups * (chunk_length * chunk_length * state_dimension)  # QK^T
    flops += num_chunks * nheads * (chunk_length * chunk_length)  # L mask
    flops += 2 * num_chunks * nheads * (chunk_length * chunk_length * head_dimension)  # V mult
    
    # Low-rank factors
    flops += num_chunks * nheads * chunk_length * 2  # decay states
    flops += 2 * (num_chunks * nheads * chunk_length * head_dimension * state_dimension)  # state computation
    
    # Inter-chunk operations
    flops += nheads * num_chunks * num_chunks * 2  # chunk decay
    flops += num_chunks * nheads * head_dimension * state_dimension  # state update
    
    # Output conversion
    flops += nheads * num_chunks * chunk_length  # decay out
    flops += 2 * (num_chunks * nheads * chunk_length * head_dimension * state_dimension)  # final output
    flops += num_chunks * nheads * chunk_length * head_dimension  # state multiply
    
    # Final addition
    flops += num_chunks * chunk_length * nheads * head_dimension
    
    return b * flops


def get_inputs_mamba(dtype: torch.dtype, b: int, h: int, n: int, dv: int, verbose: bool = True):
    """Generate inputs for Mamba-2 computation.
    
    Args:
        dtype: Data type for tensors
        b: Batch size
        h: Number of heads
        n: Sequence length
        dv: Head dimension
        verbose: Whether to print dimension info
    
    Returns:
        Tuple of input tensors and parameters
    """
    torch.manual_seed(42)
    
    batch, seqlen = b, n
    chunk_size = 64
    dim = h * dv
    headdim = dv
    ngroups = 1
    dstate = 64
    device = "cuda"
    
    if verbose:
        print(f"Dimensions: batch={batch}, seqlen={seqlen}, heads={h}, dim={dim}, headdim={headdim}")
    
    x = torch.randn(batch, seqlen, h, headdim, dtype=torch.float32, device=device)
    dt = F.softplus(torch.randn(batch, seqlen, h, dtype=torch.float32, device=device) - 4).requires_grad_()
    A = -torch.exp(torch.rand(h, dtype=torch.float32, device=device)).requires_grad_()
    B = torch.randn(batch, seqlen, ngroups, dstate, dtype=torch.float32, device=device)
    C = torch.randn(batch, seqlen, ngroups, dstate, dtype=torch.float32, device=device)
    D = torch.randn(h, dtype=torch.float32, device=device)
    
    return x, dt, A, B, C, D, chunk_size, torch.float32

class Mamba2Triton(torch.nn.Module):
    def __init__(self):
        super().__init__()
        
    def forward(self, x, dt, A, B, C, chunk_size, D=None):
        return mamba_chunk_scan_combined(x, dt, A, B, C, chunk_size, D=None)

def mamba2_test(dtype, b, h, n, dv, causal, is_forwards, method_str, num_iters=10, verbose=True, torch_compile=False, **kwargs):
    
    triton_method = Mamba2Triton()
    if torch_compile and method_str == "mamba2_triton":
        try:
            triton_method = torch.compile(triton_method)
        except Exception as e:
            print(f"Could not compile triton_method: {e}")
            
    for stage in ['warmup', 'timed']:

        start_events = [torch.cuda.Event(enable_timing=True) for _ in range(num_iters)]
        end_events = [torch.cuda.Event(enable_timing=True) for _ in range(num_iters)]
    
        for i in range(num_iters):

            x, dt, A, B, C, D, chunk_size, _ = get_inputs_mamba(dtype, b, h, n, dv, verbose)
            q = rearrange(C, "b l h d -> b h l d").to(torch.bfloat16).contiguous()
            k = rearrange(B, "b l h d -> b h l d").to(torch.bfloat16).contiguous()
            v = rearrange(x*dt.unsqueeze(-1), "b l h d -> b h l d").to(torch.bfloat16).contiguous()
            a = rearrange(A*dt, "b h l -> b l h").to(torch.float32).contiguous().requires_grad_()

            try:
                if method_str == "mamba2_tk": 
                    torch.cuda.synchronize()
                    start_events[i].record()
                    y = tk.mamba2(q, k, v, a)
                    end_events[i].record()
                    torch.cuda.synchronize()
                elif method_str == "mamba2_triton":
                    torch.cuda.synchronize()
                    start_events[i].record()
                    y = triton_method(x, dt, A, B, C, chunk_size, D=None)
                    end_events[i].record()
                    torch.cuda.synchronize()
                else:
                    assert 0, f"Unknown method: {method_str}"
            except Exception as e:
                if verbose:
                    print(f"Error: {e}")
                return None, -1

            torch.cuda.empty_cache()

    assert not np.isnan(y.detach().float().cpu()).any(), "NaN values detected in output 'o'"
    assert not np.isinf(y.detach().float().cpu()).any(), "Inf values detected in output 'o'"
    tot = sum([s.elapsed_time(e) for s, e in zip(start_events, end_events)])/num_iters
    return y, tot

    
IMPLEMENTATIONS = {
    "mamba2_triton": partial(mamba2_test, causal=True, is_forwards=True, method_str="mamba2_triton"),
    "mamba2_tk": partial(mamba2_test, causal=True, is_forwards=True, method_str="mamba2_tk"),
}

NAME = "MAMBA 2"




import torch
from tqdm import trange
import numpy as np
import sys
torch.set_grad_enabled(False)

N = 4096
B = 4
H = 1024
N1 = int(np.sqrt(N))
print(N, N1)

TESTNAME = sys.argv[1]

if TESTNAME in ['ones']:
    TESTNAME = f"ones_f{N}_H{H}_B{B}"
    u = (torch.ones((B, H, N), dtype=torch.bfloat16, device='cpu')).to(torch.float32) 
    k = (torch.ones((H, N), dtype=torch.bfloat16, device='cpu')).to(torch.float32)
    
elif TESTNAME in ['randn']:
    TESTNAME = f"randn_f{N}_H{H}_B{B}"
    torch.random.manual_seed(42)
    u = (torch.randn((B, H, N), dtype=torch.bfloat16, device='cpu')).to(torch.float32) 
    k = (torch.randn((H, N), dtype=torch.bfloat16, device='cpu')).to(torch.float32)

else:
    print('Invalid test name')
    sys.exit(0)

def ref_fftconv(u, k, N):
    L = u.shape[-1]
    u_f = torch.fft.fft(u.float(), n = N)
    k_f = torch.fft.fft(k.float(), n = N)
    y_f = u_f * k_f
    y = torch.fft.ifft(y_f, n = N).real[..., :L].to(u.dtype).contiguous()
    return y

def fft_matrix(N):
    n = torch.arange(N)
    k = n.view(-1, 1)
    M = torch.exp(-2j * torch.pi * n * k / N)
    return M

def compute_twiddle_factors_fft(n, m):
    """Compute the twiddle factors of size n x m"""
    n_a = torch.arange(n).view(-1, 1)
    m_a = torch.arange(m)
    N = n * m
    M = torch.exp(-2j * torch.pi * n_a * m_a / N)
    return M

def ifft_matrix(N):
    n = torch.arange(N)
    k = n.view(-1, 1)
    M = torch.exp(2j * torch.pi * n * k / N)
    return M

def compute_twiddle_factors_ifft(n, m):
    """Compute the twiddle factors of size n x m"""
    n_a = torch.arange(n).view(-1, 1)
    m_a = torch.arange(m)
    N = n * m
    M = torch.exp(2j * torch.pi * n_a * m_a / N)
    return M


def pytorch_test(u, k, TESTNAME='all'):
    u_real = u.to(torch.bfloat16)
    u_imag = torch.zeros_like(u, dtype=torch.bfloat16)
    f_mat = fft_matrix(N1)
    f_real = f_mat.real.to(torch.bfloat16).contiguous()
    f_imag = f_mat.imag.to(torch.bfloat16).contiguous()

    finv_mat = ifft_matrix(N1)
    finv_real = finv_mat.real.to(torch.bfloat16).contiguous()
    finv_imag = finv_mat.imag.to(torch.bfloat16).contiguous()

    # Normalization factor to make IFFT exact inverse of FFT
    tw = compute_twiddle_factors_fft(N1, N1)
    tw_real = tw.real.to(torch.bfloat16).contiguous()
    tw_imag = tw.imag.to(torch.bfloat16).contiguous()
    twinv = compute_twiddle_factors_ifft(N1, N1) / N
    twinv_real = twinv.real.to(torch.bfloat16).contiguous()
    twinv_imag = twinv.imag.to(torch.bfloat16).contiguous()

    # Compute the regular FFT if the seq len isn't 512 or 2048
    k_f = torch.fft.fft(k.float(), n = N)
    k_fT = k_f.reshape(H, N1, N1).transpose(-1, -2)
    kfT_real = k_fT.real.to(torch.bfloat16).contiguous()
    kfT_imag = k_fT.imag.to(torch.bfloat16).contiguous()

    o_real = ref_fftconv(u, k, N)
    o_real = o_real.reshape(B, H, N1, N1).to(torch.bfloat16).contiguous()

    u_real = u_real.reshape(B, H, N1, N1).to(torch.bfloat16).contiguous()
    u_imag = u_imag.reshape(B, H, N1, N1).to(torch.bfloat16).contiguous()

    return u_real, u_imag, kfT_real, kfT_imag, f_real, f_imag, finv_real, finv_imag, tw_real, tw_imag, twinv_real, twinv_imag, o_real


u_real, u_imag, kfT_real, kfT_imag, f_real, f_imag, finv_real, finv_imag, tw_real, tw_imag, twinv_real, twinv_imag, o_real = pytorch_test(u, k, TESTNAME=TESTNAME)


with open(f'{TESTNAME}.txt', 'w') as f:
    u_real_f = u_real.to(torch.float32).flatten().cpu().numpy().tolist()
    u_imag_f = u_imag.to(torch.float32).flatten().cpu().numpy().tolist()
    kfT_real_f = kfT_real.to(torch.float32).flatten().cpu().numpy().tolist()
    kfT_imag_f = kfT_imag.to(torch.float32).flatten().cpu().numpy().tolist()
    f_real_f = f_real.to(torch.float32).flatten().cpu().numpy().tolist()
    f_imag_f = f_imag.to(torch.float32).flatten().cpu().numpy().tolist()
    finv_real_f = finv_real.to(torch.float32).flatten().cpu().numpy().tolist()
    finv_imag_f = finv_imag.to(torch.float32).flatten().cpu().numpy().tolist()
    tw_real_f = tw_real.to(torch.float32).flatten().cpu().numpy().tolist()
    print(kfT_real_f[:256])
    tw_imag_f = tw_imag.to(torch.float32).flatten().cpu().numpy().tolist()
    twinv_real_f = twinv_real.to(torch.float32).reshape(-1, N1, N1).transpose(-1, -2).flatten().cpu().numpy().tolist()
    twinv_imag_f = twinv_imag.to(torch.float32).reshape(-1, N1, N1).transpose(-1, -2).flatten().cpu().numpy().tolist()
    o_real_f = o_real.to(torch.float32).flatten().cpu().numpy().tolist()

    for i in trange(len(u_real_f)):
        f.write(repr(u_real_f[i]))
        f.write(' ')
    for i in trange(len(u_imag_f)):
        f.write(repr(u_imag_f[i]))
        f.write(' ')
    for i in trange(len(kfT_real_f)):
        f.write(repr(kfT_real_f[i]))
        f.write(' ')
    for i in trange(len(kfT_imag_f)):
        f.write(repr(kfT_imag_f[i]))
        f.write(' ')
    for i in trange(len(f_real_f)):
        f.write(repr(f_real_f[i]))
        f.write(' ')
    for i in trange(len(f_imag_f)):
        f.write(repr(f_imag_f[i]))
        f.write(' ')
    for i in trange(len(finv_real_f)):
        f.write(repr(finv_real_f[i]))
        f.write(' ')
    for i in trange(len(finv_imag_f)):
        f.write(repr(finv_imag_f[i]))
        f.write(' ')
    for i in trange(len(tw_real_f)):
        f.write(repr(tw_real_f[i]))
        f.write(' ')
    for i in trange(len(tw_imag_f)):
        f.write(repr(tw_imag_f[i]))
        f.write(' ')
    for i in trange(len(twinv_real_f)):
        f.write(repr(twinv_real_f[i]))
        f.write(' ')
    for i in trange(len(twinv_imag_f)):
        f.write(repr(twinv_imag_f[i]))
        f.write(' ')
    for i in trange(len(o_real_f)):
        f.write(repr(o_real_f[i]))
        f.write(' ')
    



import torch
import thunderkittens as tk

N = 1024
hidden_states = torch.randn(N, N, device="cuda").to(torch.float8_e4m3fn).contiguous()
weights = torch.randn(N, N, device="cuda").to(torch.float8_e4m3fn).contiguous()
output = tk.fp8_gemm(hidden_states, weights)




import torch
import thunderkittens as tk
import numpy as np
from tqdm import tqdm

from flash_attn_interface import flash_attn_func

def generate_tensor(shape, mean, std, dtype, device):
    
    tensor = torch.randn(shape, dtype=dtype, device=device)
    magnitude = torch.norm(tensor, dim=-1, keepdim=True)
    scaled_tensor = tensor * (torch.randn(magnitude.shape, dtype=dtype, device=device) * std + mean) / magnitude
    
    return scaled_tensor.contiguous()


def tk_forward_test(Q, K, V, causal):
    O = torch.zeros_like(Q).contiguous()
    L = torch.zeros(Q.shape[0], Q.shape[1], Q.shape[2], 1, device=Q.device, dtype=torch.float)
    tk.mha_forward(Q, K, V, O, L, causal)
    return O, L

def tk_backward_test(Q, K, V, O, L, d_vec, dO, causal):
    
    return tk.mha_backward(Q, K, V, O, L, d_vec, dO, causal)

def check_consistency(b, h, n, d, causal, mean, std, num_iterations=100000):
    
    # Generate fixed inputs
    torch.manual_seed(0)
    Q  = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
    K  = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
    V  = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
    O  = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
    dO = generate_tensor((b, h, n, d), mean, std, torch.bfloat16, 'cuda')
    
    _, L = tk_forward_test(Q, K, V, causal)
    d_vec = torch.zeros(b, h, n, 1, device='cuda', dtype=torch.float)

    # Initial run to get reference outputs
    ref_qg, ref_kg, ref_vg = tk_backward_test(Q, K, V, O, L, d_vec, dO, causal)

    max_diff_qg, max_diff_kg, max_diff_vg = 0, 0, 0

    for _ in tqdm(range(num_iterations), desc="Checking consistency"):
        torch.cuda.synchronize()
        qg, kg, vg = tk_backward_test(Q, K, V, O, L, d_vec, dO, causal)
        torch.cuda.synchronize()
        
        max_diff_qg = torch.abs(qg - ref_qg).max()
        max_diff_kg = torch.abs(kg - ref_kg).max()
        max_diff_vg = torch.abs(vg - ref_vg).max()
        
        if max_diff_qg > 1e-7 or max_diff_kg > 1e-7 or max_diff_vg > 1e-7:
            breakpoint()

    return max_diff_qg, max_diff_kg, max_diff_vg

# Example usage
b, h, n, d = 1, 16, 768*8, 128
causal = True
mean = 1e-1
std = 1

max_diff_qg, max_diff_kg, max_diff_vg = check_consistency(b, h, n, d, causal, mean, std)


import torch 
import torch.nn.functional as F
from einops import rearrange

from baselines.ssd_minimal import ssd_minimal_discrete 

try:
    from mamba_ssm.ops.triton.ssd_combined import mamba_chunk_scan_combined
    print(f"Successfully imported mamba_chunk_scan_combined")
except ImportError:
    print(f"Failed to import mamba_chunk_scan_combined; Please 'pip install mamba_ssm'.")
    mamba_chunk_scan_combined = None


try:
    import thunderkittens as tk 
    print(f"Successfully imported thunderkittens")
except ImportError:
    print(f'Failed to import thunderkittens; Please "pip setup.py install".')
    tk = None


import warnings
warnings.filterwarnings("ignore", message=".*torch.cuda.amp.custom_bwd(args...)` is deprecated..*")
warnings.filterwarnings("ignore", message=".*torch.cuda.amp.custom_fwd(args...)` is deprecated*")




# Copyright (c) 2024, Albert Gu and Tri Dao.
"""Minimal implementation of SSD.

This is the same as Listing 1 from the paper.
"""

import torch
import torch.nn.functional as F
from einops import rearrange, repeat

from mamba_ssm.ops.triton.ssd_combined import mamba_chunk_scan_combined


# Simple test
def test_correctness():
    # torch.manual_seed(42)

    ## Dimensions
    # Denoted (B, T, Q, D, P) in the paper
    batch, nheads, seqlen, chunk_size, headdim = 16, 16, 1024, 64,  64
    ngroups = 1 # (G) in the paper
    dstate = 64  # (N) in the paper
    dtype = torch.float32
    device = "cuda"

    # TK 
    x = torch.randn(batch, seqlen, nheads, headdim, dtype=dtype, device=device)
    dt = F.softplus(torch.randn(batch, seqlen, nheads, dtype=torch.float32, device=device) - 4).requires_grad_()
    A = (-torch.exp(torch.rand(nheads, dtype=torch.float32, device=device))).requires_grad_()
    B = torch.randn(batch, seqlen, ngroups, dstate, dtype=dtype, device=device)
    C = torch.randn(batch, seqlen, ngroups, dstate, dtype=dtype, device=device)
    D = torch.randn(nheads, dtype=dtype, device=device)

    q = rearrange(C, "b l h d -> b h l d").to(torch.bfloat16).contiguous()
    k = rearrange(B, "b l h d -> b h l d").to(torch.bfloat16).contiguous()
    v = rearrange(x*dt.unsqueeze(-1), "b l h d -> b h l d").to(torch.bfloat16).contiguous()
    a = rearrange(A*dt, "b h l -> b l h").to(torch.float32).contiguous()


    for i in range(2000):
        print(f"{torch.isnan(q).any(), torch.isnan(k).any(), torch.isnan(v).any(), torch.isnan(a).any()}")
        torch.cuda.synchronize()
        y_tk = tk.mamba2(q, k, v, a) # chunk size is 64
        torch.cuda.synchronize()
        y_tk = rearrange(y_tk, "b h l d -> b l h d").contiguous().to(dtype)
        has_nan = torch.isnan(y_tk).any()
        has_inf = torch.isinf(y_tk).any()
        num_nans = torch.sum(torch.isnan(y_tk))
        print(f"Has NaN: {has_nan}; Has Inf: {has_inf}; Num of NaNs: {num_nans} / {y_tk.numel()}")
        nan_positions = torch.nonzero(torch.isnan(y_tk))
        vals_gt_1e15 = torch.nonzero(y_tk > 1e15)
        if has_nan or has_inf or vals_gt_1e15.numel() > 0:
            breakpoint()

        torch.cuda.empty_cache()

        x = torch.randn(batch, seqlen, nheads, headdim, dtype=dtype, device=device)
        dt = F.softplus(torch.randn(batch, seqlen, nheads, dtype=torch.float32, device=device) - 4).requires_grad_()
        A = (-torch.exp(torch.rand(nheads, dtype=torch.float32, device=device))).requires_grad_()
        B = torch.randn(batch, seqlen, ngroups, dstate, dtype=dtype, device=device)
        C = torch.randn(batch, seqlen, ngroups, dstate, dtype=dtype, device=device)
        D = torch.randn(nheads, dtype=dtype, device=device)
        
        q = rearrange(C, "b l h d -> b h l d").to(torch.bfloat16).contiguous()
        k = rearrange(B, "b l h d -> b h l d").to(torch.bfloat16).contiguous()
        v = rearrange(x*dt.unsqueeze(-1), "b l h d -> b h l d").to(torch.bfloat16).contiguous()
        a = rearrange(A*dt, "b h l -> b l h").to(torch.float32).contiguous()



if __name__ == "__main__":
    test_correctness()


import torch
import torch.nn.functional as F
import sys
from tqdm import tqdm


size=(4096, 4096)

TESTNAME = sys.argv[1]

if TESTNAME == 'ones':
    x = 10 * torch.ones(size, dtype=torch.float32, device='cuda')
    w = 10 * torch.ones(size, dtype=torch.float32, device='cuda').t()   # Note: cuBLASLt float8 matmul requires column major for the second argument
elif TESTNAME == 'randn':
    torch.random.manual_seed(42)
    x = torch.randn(size, dtype=torch.float32, device='cuda') * 0.2
    w = torch.randn(size, dtype=torch.float32, device='cuda').t() * 0.2
elif TESTNAME == 'custom':
    x = torch.arange(size[0] * size[1], dtype=torch.float32, device='cuda') / 100000.0
    w = x.clone()  # Since you want the same values
    x = x.reshape(size).contiguous()
    w = w.reshape(size).contiguous().t()
    print(f"x.shape: {x.shape}, w.shape: {w.shape}")
else:
    print('Invalid test name')
    sys.exit(0)

finfo = torch.finfo(torch.float8_e4m3fn)
FP8_e4m3_MAX = finfo.max
FP8_e4m3_MIN = finfo.min
print(f'FP8_e4m3_MAX: {FP8_e4m3_MAX}, FP8_e4m3_MIN: {FP8_e4m3_MIN}')

def to_float8_e4m3fn(x: torch.Tensor):
    scales = x.abs().amax(dim=-1).div(FP8_e4m3_MAX)
    x = x.div(scales.unsqueeze(-1)).clamp(min=FP8_e4m3_MIN, max=FP8_e4m3_MAX)
    x = x.to(torch.float8_e4m3fn)
    return x, scales.float()

def compare_f8_mm(dtype=torch.float8_e4m3fn) -> None:
    # do a scaled cast to float8 on the inputs
    x_f8, x_inv_s = to_float8_e4m3fn(x)
    w_f8, w_inv_s = to_float8_e4m3fn(w)

    print(f'x_inv_s: {x_inv_s[:10]}')
    print(f'w_inv_s: {w_inv_s[:10]}')

    y = torch._scaled_mm(
        x_f8, w_f8,      
        out_dtype=torch.bfloat16,
        scale_a=x_inv_s.unsqueeze(-1),     # (16, 1)
        scale_b=w_inv_s.unsqueeze(0),      # (1, 16)
        use_fast_accum=True
    ) # bias=bias

    # compare output of float8 matmul to the fp16 baseline
    cos_sim = F.cosine_similarity(torch.mm(x, w).reshape(-1), y.reshape(-1), dim=0)    
    print(f'cos_sim {cos_sim.item():.4f}')

    # average of x
    avg_x = x.abs().mean()
    print(f'avg_x {avg_x.item():.4f}')

    # average of w
    avg_w = w.abs().mean()
    print(f'avg_w {avg_w.item():.4f}')

    # average of y
    avg_y = y.abs().mean()
    print(f'avg_y {avg_y.item():.4f}')

    # max diff
    max_diff = (torch.mm(x, w) - y).abs().max()
    print(f'max_diff {max_diff.item():.4f}')

    # average diff
    avg_diff = (torch.mm(x, w) - y).abs().mean()
    print(f'avg_diff {avg_diff.item():.4f}')


if __name__ == "__main__":
    compare_f8_mm()
    






import torch
import torch.nn as nn
from tqdm import trange
import numpy as np
import sys
import math

B = 4
N = 1024
D = 64*16

TESTNAME = sys.argv[1]

norm = nn.LayerNorm(D).cuda()

if TESTNAME == 'ones':
    x = torch.zeros((B, N, D), dtype=torch.bfloat16, device='cuda').requires_grad_()
    residual = torch.ones((B, N, D), dtype=torch.bfloat16, device='cuda').requires_grad_() * 2
elif TESTNAME == 'arange':
    x = torch.zeros((B, N, D), dtype=torch.bfloat16, device='cuda')
    residual = torch.zeros((B, N, D), dtype=torch.bfloat16, device='cuda')
    for i in range(B):
        for j in range(N):
            x[i,j] = torch.arange(D, dtype=torch.bfloat16, device='cuda')
elif TESTNAME == 'randn':
    torch.random.manual_seed(42)
    x = torch.randn((B, N, D), dtype=torch.bfloat16, device='cuda').requires_grad_()
    residual = torch.randn((B, N, D), dtype=torch.bfloat16, device='cuda').requires_grad_()
else:
    print('Invalid test name')
    sys.exit(0)

def get_output(x, residual, norm, dropout_p=0.00):
    # 1. dropout on x
    mask = torch.bernoulli(torch.full_like(x, 1 - dropout_p))
    dropped = x * mask / (1 - dropout_p) 

    # 3. residual = dropped + residual
    residual = ( dropped + residual ) if residual is not None else dropped 

    # 4. norm
    mean = residual.mean(dim=-1, keepdim=True)
    var  = residual.var(dim=-1, keepdim=True, unbiased=False)
    sqrt_var = torch.sqrt(var + norm.eps)
    residual_norm = (residual - mean) / sqrt_var
    norm_weight = norm.weight
    norm_bias = norm.bias
    # norm_weight = torch.randn_like(norm_weight) # was for testing
    o = norm_weight * residual_norm + norm_bias 
    
    return o, residual, norm_weight, norm_bias, mean, sqrt_var

o, new_residual, norm_weight, norm_bias, mean, var = get_output(x, residual, norm)

with open(f'{TESTNAME}.txt', 'w') as f:
    xf = x.to(torch.float32).flatten().detach().cpu().numpy().tolist()
    residualf = residual.to(torch.float32).flatten().detach().cpu().numpy().tolist()
    norm_weightf = norm_weight.to(torch.float32).flatten().detach().cpu().numpy().tolist()
    norm_biasf = norm_bias.to(torch.float32).flatten().detach().cpu().numpy().tolist()
    meanf = mean.to(torch.float32).flatten().detach().cpu().numpy().tolist()
    varf = var.to(torch.float32).flatten().detach().cpu().numpy().tolist()
    of = o.to(torch.float32).flatten().detach().cpu().numpy().tolist()
    o_residf = new_residual.to(torch.float32).flatten().detach().cpu().numpy().tolist()

    for i in trange(B*N*D):
        f.write(repr(xf[i]))
        f.write(' ')

    for i in trange(B*N*D):
        f.write(repr(residualf[i]))
        f.write(' ')

    for i in trange(B*N*D):
        f.write(repr(of[i]))
        if (i < 5): print(of[i]) 
        f.write(' ')

    for i in trange(B*N*D):
        f.write(repr(o_residf[i]))
        f.write(' ')

    for i in trange(D):
        f.write(repr(norm_weightf[i]))
        f.write(' ')
    
    for i in trange(D):
        f.write(repr(norm_biasf[i]))
        f.write(' ')
    
    for i in trange(B*N*1):
        f.write(repr(meanf[i]))
        f.write(' ')

    for i in trange(B*N*1):
        f.write(repr(varf[i]))
        f.write(' ')




"""
General helper functions for setting up experiments
"""
import os
import random

from argparse import ArgumentParser
from omegaconf import DictConfig

import torch
import numpy as np

from .logging import _format_arg


def init_wandb(args: ArgumentParser) -> any:
    """Initialize WandB"""
    if args.no_wandb:
        wandb = None
    else:
        import wandb
        wandb.init(config={},
                   entity=args.wandb_entity,
                   name=args.run_name,
                   project=args.project_name)
    return wandb


def seed_everything(seed: int) -> None:
    """
    Seed everything
    """
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def get_run_name_from_checkpoint(checkpoint_path: str) -> str:
    """
    Helper function to get a condensed run name from the checkpoint path
    """
    name = []
    for s in checkpoint_path.split('/')[-1].split('-'):
        if '.pt' in s:
            name.append(f'_{s[:-3]}')
        try:
            s = s.split('=')
            s = ''.join([c[0] for c in s[1].split('_')])
            name.append(s)
        except IndexError:
            pass
    return ''.join(name)


def get_run_name_from_args(args) -> str:
    """
    Prepare a heinous identifier for the run based on args
    """
    if args.load_distill_checkpoint is not None and args.load_distill_checkpoint != 'default':
        distill_name = get_run_name_from_checkpoint(args.load_distill_checkpoint)
    else:
        distill_name = args.distill_config
    if args.load_finetune_checkpoint is not None and args.finetune_config is None:  # args.load_finetune_checkpoint != 'default':
        finetune_name = get_run_name_from_checkpoint(args.load_finetune_checkpoint)
    else:
        finetune_name = args.finetune_config
    args.run_name = f'dl-d={distill_name}-m={args.model_config}-f={finetune_name}'
    if args.no_peft_grad_ckpt is not None:
        args.run_name += f'-npgc={args.no_peft_grad_ckpt}'
    args.run_name += f'-s={args.seed}'
    if args.debug:
        args.run_name += f'-debug'
    if args.no_attention_mask is not None:
        args.run_name += f'-nam=1'
    return args.run_name.replace('True', '1').replace('False', '0')  # concise hacks


def flatten_config(config: dict, flattened: dict, key: str) -> dict:
    """
    Recursive way to flatten config args for saving to WandB
    """
    for k, v in config.items():
        if isinstance(v, dict):
            flatten_config(v, flattened, f'{key}{k}_')
        elif isinstance(v, list):
            for ix, _config in enumerate(v):
                if isinstance(_config, dict):
                    flatten_config(_config, flattened, f'{key}{k}_{ix}_')
        else:
            flattened[f'{key}{k}'] = v
    return flattened


def update_config_from_args(config: DictConfig,
                            args: ArgumentParser,
                            ignore_args: list = None) -> DictConfig:
    """
    Quick hacks to override default configs
    """
    ignore_args = [] if ignore_args is None else ignore_args
    
    # Dataset
    if getattr(args, 'dataset', None):
        config.dataset.name = args.dataset
        args.run_name += f'-ds={args.dataset}'
    
    # Optimizer
    for arg in ['lr', 'weight_decay']:
        if arg not in ignore_args:
            argval = getattr(args, arg, None)
            if argval is not None:
                setattr(config.optimizer, arg, argval)
                args.run_name += f'-{_format_arg(arg)}={argval}'
    try:
        if getattr(args, 'optim', None):
            config.optimizer.optim = args.optim
            args.run_name += f'-o={args.optim}'
    except AttributeError:
        pass
    
    # Scheduler
    try:
        if getattr(args, 'scheduler', None):
            config.lr_scheduler.lr_scheduler_type = args.scheduler
            args.run_name += f'-sc={args.scheduler}'
    except AttributeError:
        pass

    # Dataset
    for arg in [a for a in dir(args) if 'dataset_' in a]:
        argval = getattr(args, arg, None)
        if argval is not None:
            setattr(config.dataset.dataset_config, arg[len('dataset_'):], argval)
            args.run_name += f'-{_format_arg(arg)}={argval}'

    # Dataloader
    for arg in ['batch_size']:  # , 'num_workers']:
        argval = getattr(args, arg, None)
        if argval is not None:
            setattr(config.dataloader, arg, argval)
            args.run_name += f'-{_format_arg(arg)}={argval}'

    # Trainer
    for arg in ['gradient_accumulation_steps', 'num_train_epochs', 
                'max_steps', 'max_finetune_steps', 'eval_steps', 
                'seed', 'max_eval_batches']:
        argval = getattr(args, arg, None)
        if argval is not None:
            setattr(config.trainer, arg, argval)
            if arg in ['max_steps', 'max_finetune_steps',
                       'gradient_accumulation_steps', 'num_train_epochs', 'seed']:
                args.run_name += f'-{_format_arg(arg)}={argval}'

    # Misc
    for arg in ['replicate']:
        argval = getattr(args, arg, None)
        if argval is not None:
            args.run_name += f'-{_format_arg(arg)}={argval}'

    return config


def update_model_config_from_args(model_config: DictConfig, 
                                  args: ArgumentParser) -> DictConfig:
    """
    Override default configs given argparse args
    """
    # Overall attention 
    for arg in ['attention_type', 'learned_kernel', 'tie_qk_kernels',
                'train_qk', 'state_chunk_len', 'no_peft_grad_ckpt']:
        argval = getattr(args, arg, None)
        if argval is not None:
            setattr(model_config['attention'], arg, argval)
            args.run_name += f'-{_format_arg(arg)}={argval}'
        else:
            try:
                getattr(model_config['attention'], arg)
            except AttributeError:
                setattr(model_config['attention'], arg, None)

    # Learned kernel
    for arg in ['lk_skip_connection', 'lk_zero_init', 'lk_normal_init']:
        argval = getattr(args, arg, None)
        if argval is not None:
            setattr(model_config['attention']['learned_kernel_kwargs'], 
                    arg[len('lk_'):], argval)
            args.run_name += f'-{_format_arg(arg)}={argval}'
            
    # Pretrained model
    if args.pretrained_model_name_or_path is not None:  # if specified 
        pmnop = args.pretrained_model_name_or_path
        model_config.model.pretrained_model_name_or_path = pmnop
        args.run_name += f'-pmnop={pmnop.split("/")[-1]}'
        
    return model_config



import os    
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat
from functools import partial

try:
    import thunderkittens as tk
    print(f"Successfully imported thunderkittens")
except:
    print("Could not import thunderkittens")

try:
    from layernorm.baselines.layer_norm_triton import layer_norm_fn, RMSNorm
    print(f"Successfully imported layer_norm_fn")
except:
    layer_norm_fn = None
    print("Could not import layer_norm_fn. Please obtain from FlashAttention repo.")


################## Layer norm ######################

def get_flops(batch, seqlen, headdim, nheads):
    f = batch*seqlen*nheads*headdim # compute the mask for dropout 
    f += batch*seqlen*nheads*headdim # add dropout and residual
    f += batch*seqlen*nheads*headdim # compute the mean
    f += batch*seqlen*nheads*headdim # compute the variance
    f += batch*seqlen*nheads*headdim # subtract mean
    f += batch*seqlen*nheads*headdim # divide by variance
    f += batch*seqlen*nheads*headdim # multiply by norm weight 
    f += batch*seqlen*nheads*headdim # add norm bias
    return f


def get_layer_norm_inputs(b, h, n, dv, dt):
    d_model = h * dv
    p = 0.1

    norm = nn.LayerNorm(d_model).cuda()
    dropout = nn.Dropout(p)

    x = torch.randn(b, n, d_model).to(dtype=dt, device='cuda')
    residual = torch.randn(b, n, d_model).to(dtype=dt, device='cuda')
    norm_weight = norm.weight.to(dtype=torch.bfloat16)
    norm_bias = norm.bias.to(dtype=torch.bfloat16)
    out = torch.zeros_like(x)
    out_resid = torch.zeros_like(x)
    return x, residual, norm_weight, norm_bias, out, out_resid, norm, dropout

class pytorch_layernorm(torch.nn.Module):
    def __init__(self, d_model, p):
        super().__init__()
        self.dropout = nn.Dropout(p).cuda()
        self.norm = nn.LayerNorm(d_model).cuda()
        
    def forward(self, x, residual, norm_weight, norm_bias, dropout_p):
        with torch.no_grad():
            dropped = self.dropout(x) 
            residual = (residual + dropped ) if residual is not None else dropped
            y = self.norm(residual.to(dtype=self.norm.weight.dtype))
            residual = residual.to(torch.float32)
        return y, residual
        
def layernorm_test(dt, b, h, n, dv, causal, is_forwards, method_str, num_iters=10, verbose=True, torch_compile=True, **kwargs):
    
    pytorch_method = pytorch_layernorm(d_model=h*dv, p=0.1)
    if torch_compile and method_str == "pytorch_layernorm":
        try:
            pytorch_method = torch.compile(pytorch_method)
        except Exception as e:
            print(f"Could not compile pytorch_layernorm: {e}")
                     
    for stage in ['warmup', 'timed']:
        start_events = [torch.cuda.Event(enable_timing=True) for _ in range(num_iters)]
        end_events = [torch.cuda.Event(enable_timing=True) for _ in range(num_iters)]
    
        for i in range(num_iters):
            (
                x, residual, norm_weight, norm_bias, 
                out, out_resid, norm, dropout 
            ) = get_layer_norm_inputs(b, h, n, dv, dt)
            rowscale = torch.ones(x.shape[:-1], device=x.device, dtype=x.dtype, )
            residual_in_fp32 = True
            dtype = dt
            dropout_p = dropout.p
                
            if True:
                if method_str == 'pytorch_layernorm':
                    torch.cuda.synchronize()
                    start_events[i].record()
                    y, residual = pytorch_method(x, residual, norm_weight, norm_bias, dropout_p)
                    end_events[i].record()
                    torch.cuda.synchronize()

                elif method_str == 'triton_layernorm':
                    torch.cuda.synchronize()
                    start_events[i].record()
                    with torch.no_grad():
                        y, residual = layer_norm_fn(
                            x,
                            norm.weight,
                            norm.bias,
                            residual=residual,
                            eps=norm.eps,
                            dropout_p=dropout.p,
                            rowscale=rowscale,
                            prenorm=True,
                            residual_in_fp32=residual_in_fp32,
                            is_rms_norm=False
                        )
                    end_events[i].record()
                    torch.cuda.synchronize()
                
                elif method_str == 'tk_layernorm': 
                    torch.cuda.synchronize()
                    start_events[i].record()
                    with torch.no_grad():
                        y, residual = tk.fused_layernorm(
                            x, residual, 
                            norm_weight, norm_bias, 
                            dropout.p,
                        )
                    end_events[i].record()
                    torch.cuda.synchronize()

                else:
                    raise ValueError(f"Unknown method: {method_str}")

            # except Exception as e:
            #     if verbose:
            #         print(f"Error: {e}")
            #     return None, -1

            torch.cuda.empty_cache()

    assert not np.isnan(y.detach().float().cpu()).any(), "NaN values detected in output 'o'"
    assert not np.isinf(y.detach().float().cpu()).any(), "Inf values detected in output 'o'"
    tot = sum([s.elapsed_time(e) for s, e in zip(start_events, end_events)])/num_iters
    return y, tot


IMPLEMENTATIONS = {
    'pytorch_layernorm': partial(layernorm_test, causal=True, is_forwards=True, method_str='pytorch_layernorm'),
    'triton_layernorm': partial(layernorm_test, causal=True, is_forwards=True, method_str='triton_layernorm'),
    'tk_layernorm': partial(layernorm_test, causal=True, is_forwards=True, method_str='tk_layernorm'),
}

NAME = "LAYER NORM"



import torch
from tqdm import trange
import numpy as np
import sys
from einops import rearrange, repeat
import math

# only generate a single batch/head of data, which makes file loading much faster.
# it does mean we'll have to check batch/head behavior separately later, but that should be much easier to debug.
B = 1
H = 1
N = 2048 if len(sys.argv) <= 2 else int(sys.argv[2])
D = 128 if len(sys.argv) <= 3 else int(sys.argv[3])

softmax_scale = 1 / math.sqrt(D)

TESTNAME = sys.argv[1]

if TESTNAME == 'ones':
    q           = torch.ones((B, H, N, D), dtype=torch.bfloat16, device='cuda')
    k           = torch.ones((B, H, N, D), dtype=torch.bfloat16, device='cuda')
    v           = torch.ones((B, H, N, D), dtype=torch.bfloat16, device='cuda')
    grad_output = torch.ones((B, H, N, D), dtype=torch.bfloat16, device='cuda')
elif TESTNAME == 'randn':
    torch.random.manual_seed(42)
    q           = torch.randn((B, H, N, D), dtype=torch.bfloat16, device='cuda')
    k           = torch.randn((B, H, N, D), dtype=torch.bfloat16, device='cuda')
    v           = torch.randn((B, H, N, D), dtype=torch.bfloat16, device='cuda')
    grad_output = torch.randn((B, H, N, D), dtype=torch.bfloat16, device='cuda')
elif TESTNAME == 'qk_test':
    q = torch.eye(D).reshape((1,1,D,D)).repeat(B, H, N//D, 1)*10
    q = q.to(dtype=torch.bfloat16, device='cuda')
    
    k = torch.eye(D).reshape((1,1,D,D)).repeat(B, H, N//D, 1)*10
    k = k.to(dtype=torch.bfloat16, device='cuda')
    
    v = torch.eye(D).reshape((1,1,D,D)).repeat(B, H, N//D, 1)*10
    v = v.to(dtype=torch.bfloat16, device='cuda')
    
    grad_output = torch.eye(D).reshape((1,1,D,D)).repeat(B, H, N//D, 1)*10
    grad_output = grad_output.to(dtype=torch.bfloat16, device='cuda')
elif TESTNAME == 'v_orientation':
    q = torch.ones((B, H, N, D), dtype=torch.bfloat16, device='cuda')
    k = torch.ones((B, H, N, D), dtype=torch.bfloat16, device='cuda')
    v = (torch.arange(D, dtype=torch.bfloat16, device='cuda')/D).reshape((1,1,1,-1)).repeat(B, H, N, 1)
    
    grad_output = torch.ones((B, H, N, D), dtype=torch.bfloat16, device='cuda')
else:
    print('Invalid test name')
    sys.exit(0)

o = torch.nn.functional.scaled_dot_product_attention(q, k, v, is_causal=False)

fn = f'{TESTNAME}_{N}_{D}.txt'
with open(fn, 'w') as f:
    # inputs
    qf = q.transpose(1,2).to(torch.float32).flatten().detach().cpu().numpy() #
    kf = k.transpose(1,2).to(torch.float32).flatten().detach().cpu().numpy()
    vf = v.transpose(1,2).to(torch.float32).flatten().detach().cpu().numpy()
    of = o.transpose(1,2).to(torch.float32).flatten().detach().cpu().numpy()
    
    for i in trange(B*H*N*D):
        f.write(repr(qf[i]))
        f.write(' ')
    for i in trange(B*H*N*D):
        f.write(repr(kf[i]))
        f.write(' ')
    for i in trange(B*H*N*D):
        f.write(repr(vf[i]))
        f.write(' ')
    for i in trange(B*H*N*D):
        f.write(repr(of[i]))
        f.write(' ')


import torch
import torch.nn.functional as F
from einops import rearrange, repeat

from tqdm import trange
import numpy as np
import sys
import math

def segsum(x):
    """More stable segment sum calculation."""
    T = x.size(-1)
    x = repeat(x, "... d -> ... d e", e=T)
    mask = torch.tril(torch.ones(T, T, device=x.device, dtype=bool), diagonal=-1)
    x = x.masked_fill(~mask, 0)
    x_segsum = torch.cumsum(x, dim=-2)
    mask = torch.tril(torch.ones(T, T, device=x.device, dtype=bool), diagonal=0)
    x_segsum = x_segsum.masked_fill(~mask, -torch.inf)
    return x_segsum

def ssd_minimal_discrete(X, A, B, C, block_len, initial_states=None):
    """
    Arguments:
        X: (batch, length, n_heads, d_head)
        A: (batch, length, n_heads)
        B: (batch, length, n_heads, d_state)
        C: (batch, length, n_heads, d_state)
    Return:
        Y: (batch, length, n_heads, d_head)
    """
    assert X.dtype == A.dtype == B.dtype == C.dtype
    assert X.shape[1] % block_len == 0

    # Rearrange into blocks/chunks
    X, A, B, C = [rearrange(x, "b (c l) ... -> b c l ...", l=block_len) for x in (X, A, B, C)]
    
    A = rearrange(A, "b c l h -> b h c l")
    A_cumsum = torch.cumsum(A, dim=-1)

    # 1. Compute the output for each intra-chunk (diagonal blocks)
    L = torch.exp(segsum(A))
    Y_diag  = torch.einsum("bclhn,bcshn,bhcls,bcshp->bclhp", C, B, L, X)

    # 2. Compute the state for each intra-chunk
    # (right term of low-rank factorization of off-diagonal blocks; B terms)
    decay_states = torch.exp((A_cumsum[:, :, :, -1:] - A_cumsum))
    states = torch.einsum("bclhn,bhcl,bclhp->bchpn", B, decay_states, X)

    # 3. Compute the inter-chunk SSM recurrence; produces correct SSM states at chunk boundaries
    # (middle term of factorization of off-diag blocks; A terms)
    if initial_states is None:
        initial_states = torch.zeros_like(states[:, :1])
    states = torch.cat([initial_states, states], dim=1)
    
    decay_chunk = torch.exp(segsum(F.pad(A_cumsum[:, :, :, -1], (1, 0))))
    
    new_states = torch.einsum("bhzc,bchpn->bzhpn", decay_chunk, states)
    states, final_state = new_states[:, :-1], new_states[:, -1]

    # 4. Compute state -> output conversion per chunk
    # (left term of low-rank factorization of off-diagonal blocks; C terms)
    state_decay_out = torch.exp(A_cumsum)
    Y_off = torch.einsum('bclhn,bchpn,bhcl->bclhp', C, states, state_decay_out)

    # Add output of intra-chunk and inter-chunk terms (diagonal and off-diagonal blocks)
    Y = rearrange(Y_diag+Y_off, "b c l h p -> b (c l) h p")
    
    return Y, final_state

torch.manual_seed(0)

batch = 16
n_heads = 16

length     = 1024 # 2048 # int(sys.argv[1])
chunk_size = 64
num_chunks = length // chunk_size

d_head  = 64
d_state = 64
ngroups = 1

# assert batch == 1   # expected by c++ harness loader
# assert n_heads == 1 # expected by c++ harness loader

# print hardcoded values
print("-----------------")
print("batch:", batch)
print("n_heads:", n_heads)
print("ngroups:", ngroups)
print("length:", length)
print("d_head:", d_head)
print("d_state:", d_state)
print("-----------------")

dtype  = torch.float32
device = 'cuda'

test_segsum = torch.arange(4, dtype=dtype, device=device)
test_segsum = segsum(test_segsum)

x  = torch.randn(batch, length, n_heads, d_head, dtype=dtype, device=device)
dt = F.softplus(torch.randn(batch, length, n_heads, dtype=torch.float32, device=device) - 4).requires_grad_()
A = (-torch.exp(torch.rand(n_heads, dtype=torch.float32, device=device))).requires_grad_()
B = torch.randn(batch, length, ngroups, d_state, dtype=dtype, device=device)
C = torch.randn(batch, length, ngroups, d_state, dtype=dtype, device=device)

X = x*dt.unsqueeze(-1)
A = A*dt
B = B
C = C

y_ref, _ = ssd_minimal_discrete(X, A, B, C, block_len=chunk_size)

X = X.permute(0, 2, 1, 3)
A = A.permute(0, 2, 1)
B = B.permute(0, 2, 1, 3)
C = C.permute(0, 2, 1, 3)
y_ref = y_ref.permute(0, 2, 1, 3)


# print shapes 
print("X:", X.shape)
print("A:", A.shape)
print("B:", B.shape)
print("C:", C.shape)
print("Y:", y_ref.shape)
print("-----------------")

filename = f'randn_{length}N_{d_head}D_{d_state}S.txt'

with open(filename, 'w') as f:
    
    # inputs
    xf = X.to(torch.float32).flatten().detach().cpu().numpy()
    Af = A.to(torch.float32).flatten().detach().cpu().numpy()
    Bf = B.to(torch.float32).flatten().detach().cpu().numpy()
    Cf = C.to(torch.float32).flatten().detach().cpu().numpy()
    
    yf = y_ref.to(torch.float32).flatten().detach().cpu().numpy()
    
    print(f'X (V) shape: {X.shape}, total elements {X.shape[0] * X.shape[1] * X.shape[2] * X.shape[3]}')
    print(f'A (alpha) shape: {A.shape}, total elements {A.shape[0] * A.shape[1] * A.shape[2]}')
    print(f'B (K) shape: {B.shape}, total elements {B.shape[0] * B.shape[1] * B.shape[2] * B.shape[3]}')
    print(f'C (Q) shape: {C.shape}, total elements {C.shape[0] * C.shape[1] * C.shape[2] * C.shape[3]}')
    print(f'Y (O) shape: {y_ref.shape}, total elements {y_ref.shape[0] * y_ref.shape[1] * y_ref.shape[2] * y_ref.shape[3]}')

    print((C[0,0]@B[0,0].T * torch.exp(segsum(A[0,0]))) @ X[0,0])
    print(y_ref[0,0])
    # print(torch.exp(segsum(A[0,0])))

    for i in trange(X.shape[0] * X.shape[1] * X.shape[2] * X.shape[3]):
        f.write(repr(xf[i])) # V
        f.write(' ')
    for i in trange(A.shape[0] * A.shape[1] * A.shape[2]):
        f.write(repr(Af[i])) # A
        f.write(' ')
    for i in trange(B.shape[0] * B.shape[1] * B.shape[2] * B.shape[3]):
        f.write(repr(Bf[i])) # K
        f.write(' ')
    for i in trange(C.shape[0] * C.shape[1] * C.shape[2] * C.shape[3]):
        f.write(repr(Cf[i])) # Q
        f.write(' ')
    for i in trange(y_ref.shape[0] * y_ref.shape[1] * y_ref.shape[2] * y_ref.shape[3]):
        f.write(repr(yf[i])) # O
        f.write(' ')











""" Source: https://github.com/state-spaces/mamba/blob/main/mamba_ssm/utils/hf.py """

import json
import torch
from transformers.utils import WEIGHTS_NAME, CONFIG_NAME
from transformers.utils.hub import cached_file


def load_config_hf(model_name):
    resolved_archive_file = cached_file(model_name, CONFIG_NAME, _raise_exceptions_for_missing_entries=False)
    return json.load(open(resolved_archive_file))

def load_state_dict_hf(model_name, device=None, dtype=None):
    # If not fp32, then we don't want to load directly to the GPU
    mapped_device = "cpu" if dtype not in [torch.float32, None] else device
    resolved_archive_file = cached_file(model_name, WEIGHTS_NAME, _raise_exceptions_for_missing_entries=False)
    return torch.load(resolved_archive_file, map_location=mapped_device)




import torch
from functools import partial
import numpy as np
import thunderkittens as tk

try:
    # To compare to Flash Attention 2
    from flash_attn import flash_attn_func as flash_attn_func2
    print(f"Successfully imported flash_attention")
except:
    flash_attn_func = None
    print("Could not import flash_attention. Please: pip install flash-attn")

try:
    # To compare to Flash Attention 3
    from flash_attn_interface import flash_attn_func as flash_attn_func3
    print(f"Successfully imported flash_attn_interface")
except:
    flash_attn_func3 = None
    print("Could not import flash_attn_interface. Please: clone flash-attention and install the hopper kernels")


def get_flops(batch, seqlen, nheads, headdim, causal, mode="fwd"):
    assert mode in ["fwd", "bwd", "fwd_bwd"]
    f = 4 * batch * seqlen**2 * nheads * headdim // (2 if causal else 1)
    flops = f if mode == "fwd" else (2.5 * f if mode == "bwd" else 3.5 * f)
    return flops


################ ATTENTION ################

def get_attention_inputs(b, h, n, dv, dt=torch.bfloat16, pad_multiple=0, ):

    if pad_multiple:
        n = (n // pad_multiple) * pad_multiple + pad_multiple  

    q = torch.randn(b, h, n, dv, dtype=dt, device='cuda').requires_grad_(True)
    k = torch.randn(b, h, n, dv, dtype=dt, device='cuda').requires_grad_(True)
    v = torch.randn(b, h, n, dv, dtype=dt, device='cuda').requires_grad_(True)
    dO = torch.randn(b, h, n, dv, dtype=dt, device='cuda')
    return q, k, v, dO

class pytorch_attention(torch.nn.Module):
    def __init__(self):
        super().__init__()
    def forward(self, q, k, v, causal):
        QK = torch.matmul(q, k.transpose(-2, -1))
        QK /= (q.size(-1) ** 0.5)
        if causal:
            mask = torch.triu(torch.ones(QK.size(-2), QK.size(-1)), 1).to(torch.bool).to(QK.device)
            QK.masked_fill_(mask, float('-inf'))
        QK = torch.nn.functional.softmax(QK, dim=-1)
        y = torch.matmul(QK, v)
        return y

def attention_test(dt, b, h, n, dv, causal, is_forwards, method_str, num_iters=10, verbose=True, torch_compile=False, **kwargs):
    
    pytorch_method = pytorch_attention()
    if torch_compile and method_str == "pytorch":
        try:
            pytorch_method = torch.compile(pytorch_method)
        except Exception as e:
            print(f"Could not compile pytorch_method: {e}")
                    
    for stage in ['warmup', 'timed']:
        start_events = [torch.cuda.Event(enable_timing=True) for _ in range(num_iters)]
        end_events = [torch.cuda.Event(enable_timing=True) for _ in range(num_iters)]

        for i in range(num_iters):
            try:
                q, k, v, dO = get_attention_inputs(b, h, n, dv, dt)
                
                if method_str == "pytorch":
                    torch.cuda.synchronize()
                    start_events[i].record()
                    y = pytorch_method(q, k, v, causal)
                    end_events[i].record()
                    torch.cuda.synchronize()

                elif method_str == "fa2": 
                    torch.cuda.synchronize()
                    start_events[i].record()
                    y = torch.nn.functional.scaled_dot_product_attention(q, k, v, is_causal=causal)
                    end_events[i].record()
                    torch.cuda.synchronize()

                elif method_str == 'tk':
                    torch.cuda.synchronize()
                    start_events[i].record()
                    y, l_vec = tk.mha_forward(q, k, v, causal)
                    end_events[i].record()
                    torch.cuda.synchronize()

                elif method_str == "fa3":
                    q = q.transpose(1,2)
                    k = k.transpose(1,2)
                    v = v.transpose(1,2)
                    dO = dO.transpose(1,2)
                    torch.cuda.synchronize()
                    start_events[i].record()
                    y, lse = flash_attn_func3(q, k, v, causal=causal)
                    end_events[i].record() 
                    torch.cuda.synchronize()

                else:
                    assert 0, f"Unknown method: {method_str}"

                outputs = ( y )

            except Exception as e:
                if verbose:
                    print(f"Error: {e}")
                return None, -1

            if is_forwards and stage == 'timed':
                continue

            try:

                if method_str == "pytorch":
                    torch.cuda.synchronize()
                    start_events[i].record()
                    y.backward(dO)
                    q_grad = q.grad.to(torch.bfloat16)
                    k_grad = k.grad.to(torch.bfloat16)
                    v_grad = v.grad.to(torch.bfloat16)
                    y = y.to(torch.bfloat16)
                    end_events[i].record()
                    torch.cuda.synchronize()

                elif method_str == "fa2":
                    torch.cuda.synchronize()
                    start_events[i].record()
                    y.backward(dO)
                    q_grad = q.grad
                    k_grad = k.grad
                    v_grad = v.grad
                    end_events[i].record()
                    torch.cuda.synchronize()

                elif method_str == 'tk':
                    torch.cuda.synchronize()
                    start_events[i].record()
                    q_grad, k_grad, v_grad = tk.mha_backward(q, k, v, y, l_vec, dO, causal)
                    end_events[i].record()
                    torch.cuda.synchronize()

                elif method_str == "fa3":
                    torch.cuda.synchronize()
                    start_events[i].record()
                    y.backward(dO)
                    q_grad = q.grad
                    k_grad = k.grad
                    v_grad = v.grad
                    end_events[i].record()
                    torch.cuda.synchronize()

                else:
                    assert 0, f"Unknown method: {method_str}"

                outputs = (q_grad, k_grad, v_grad)

            except Exception as e:
                if verbose:
                    print(f"Error: {e}")
                return (None, None, None), -1

            torch.cuda.empty_cache()
    
    tot = sum([s.elapsed_time(e) for s, e in zip(start_events, end_events)])/num_iters
    # for output in outputs:
    #     try:
    #         assert not np.isnan(output.detach().float().cpu()).any(), "NaN values detected in output 'o'"
    #         assert not np.isinf(output.detach().float().cpu()).any(), "Inf values detected in output 'o'"
    #     except: 
    #         breakpoint()
    return outputs, tot


IMPLEMENTATIONS = {
    "attn_fwd_PT_c=t": partial(attention_test, causal=True, is_forwards=True, method_str="pytorch"),
    "attn_fwd_FA2_c=t": partial(attention_test, causal=True, is_forwards=True, method_str="fa2"),
    "attn_fwd_FA3_c=t": partial(attention_test, causal=True, is_forwards=True, method_str="fa3"),
    "attn_fwd_TK_c=t": partial(attention_test, causal=True, is_forwards=True, method_str="tk"),
    "attn_fwd_PT_c=f": partial(attention_test, causal=False, is_forwards=True, method_str="pytorch"),
    "attn_fwd_FA2_c=f": partial(attention_test, causal=False, is_forwards=True, method_str="fa2"),
    "attn_fwd_FA3_c=f": partial(attention_test, causal=False, is_forwards=True, method_str="fa3"),
    "attn_fwd_TK_c=f": partial(attention_test, causal=False, is_forwards=True, method_str="tk"),
}

IMPLEMENTATIONS_BWD = {
    "attn_bwd_PT_c=t": partial(attention_test, causal=True, is_forwards=False, method_str="pytorch"),
    "attn_bwd_FA2_c=t": partial(attention_test, causal=True, is_forwards=False, method_str="fa2"),
    "attn_bwd_FA3_c=t": partial(attention_test, causal=True, is_forwards=False, method_str="fa3"),
    "attn_bwd_TK_c=t": partial(attention_test, causal=True, is_forwards=False, method_str="tk"),
    "attn_bwd_PT_c=f": partial(attention_test, causal=False, is_forwards=False, method_str="pytorch"),
    "attn_bwd_FA2_c=f": partial(attention_test, causal=False, is_forwards=False, method_str="fa2"),
    "attn_bwd_FA3_c=f": partial(attention_test, causal=False, is_forwards=False, method_str="fa3"),
    "attn_bwd_TK_c=f": partial(attention_test, causal=False, is_forwards=False, method_str="tk"),
}

NAME = "ATTENTION"




import torch 
import time 
import torch.nn as nn
import matplotlib.pyplot as plt
import os

import thunderkittens as tk

warmup_iters = 10
num_iters = 100


def run_torch(x, residual, drop_path, dropout, norm, residual_in_fp32=False):
    timings = []

    for i in range(num_iters):
        torch.cuda.synchronize()
        st = time.time()

        # dropped = drop_path(dropout(x))
        dropped = dropout(x)
        residual = (residual + dropped) if residual is not None else dropped
        x = norm(residual.to(dtype=norm.weight.dtype))
        residual = residual.to(torch.float32)

        torch.cuda.synchronize()
        et = time.time()
        tot = (et - st)
        if i > warmup_iters:
            timings.append(tot)

    tot = sum(timings)/len(timings)
    return tot


def run_flash(x, residual, drop_path, dropout, norm, residual_in_fp32=True):
    from baselines.layer_norm_triton import layer_norm_fn, RMSNorm

    timings = []

    for i in range(num_iters):

        torch.cuda.synchronize()
        st = time.time()
        rowscale = torch.ones(
                x.shape[:-1],
                device=x.device,
                dtype=x.dtype,
            )
        # drop_path(
            
        # )
        x, residual = layer_norm_fn(
            x,
            norm.weight,
            norm.bias,
            residual=residual,
            eps=norm.eps,
            dropout_p=dropout.p,
            rowscale=rowscale,
            prenorm=True,
            residual_in_fp32=residual_in_fp32,
            is_rms_norm=False
        )
        torch.cuda.synchronize()
        et = time.time()
        tot = (et - st)
        if i > warmup_iters:
            timings.append(tot)

    tot = sum(timings)/len(timings)
    return tot


def run_tk(x, residual, drop_path, dropout, norm, residual_in_fp32=False):
    x = x.to(dtype=torch.bfloat16)
    residual = residual.to(dtype=torch.bfloat16)
    norm_weight = norm.weight.to(dtype=torch.bfloat16)
    norm_bias = norm.bias.to(dtype=torch.bfloat16)

    has_residual = int(residual is not None)
    out = torch.zeros_like(x)
    out_resid = torch.zeros_like(x)

    timings = []

    for i in range(num_iters):

        torch.cuda.synchronize()
        st = time.time()
        
        out, out_resid = tk.fused_layernorm(
            x, residual, 
            norm_weight, norm_bias, 
            0.1
        )

        torch.cuda.synchronize()
        et = time.time()
        tot = (et - st)
        if i > warmup_iters:
            timings.append(tot)

    tot = sum(timings)/len(timings)
    return tot


def plot_timings(x_values, y_torch, y_flash, y_tk, x_label, y_label, title, filename):
    plt.figure(figsize=(10, 6))
    plt.plot(x_values, y_torch, marker='o', label='Torch')
    plt.plot(x_values, y_flash, marker='s', label='Flash')
    plt.plot(x_values, y_tk, marker='o', label='TK')
    plt.xlabel(x_label, fontsize=14)
    plt.ylabel(y_label, fontsize=14)
    plt.title(title, fontsize=14)
    plt.legend(fontsize=14)
    plt.grid(True)
    plt.show()

    plt.savefig(f"plots/{filename}")  # Save the plot
    plt.show()


if __name__ == "__main__":
    p = 0.1

    if not os.path.exists("plots"):
        os.makedirs("plots/")

    # Varying batch size
    n, d = 2048, 1024
    batch_sizes = [1, 2, 4, 8, 16, 32]
    torch_timings_batch = []
    flash_timings_batch = []
    tk_timings_batch = []
    
    for b in batch_sizes:
        torch.manual_seed(0)
        x = torch.randn((b, n, d), device='cuda')
        residual = torch.randn((b, n, d), device='cuda')
        norm = nn.LayerNorm(d).cuda()
        dropout = nn.Dropout(p)
        drop_path = None #StochasticDepth(p, mode="row")
        torch_timings_batch.append(run_torch(x, residual, drop_path, dropout, norm))
        flash_timings_batch.append(run_flash(x, residual, drop_path, dropout, norm))
        tk_timings_batch.append(run_tk(x, residual, drop_path, dropout, norm))
    
    plot_timings(batch_sizes, torch_timings_batch, flash_timings_batch, tk_timings_batch, 
                 'Batch Size', 'Time (s)', 'Performance vs Batch Size', 'batch_size')

    # Varying sequence length
    b, d = 16, 1024
    seq_lengths = [512, 1024, 2048, 4096]
    torch_timings_seq = []
    flash_timings_seq = []
    tk_timings_seq = []
    
    for n in seq_lengths:
        torch.manual_seed(0)
        x = torch.randn((b, n, d), device='cuda')
        residual = torch.randn((b, n, d), device='cuda')
        norm = nn.LayerNorm(d).cuda()
        dropout = nn.Dropout(p)
        drop_path = None #StochasticDepth(p, mode="row")
        torch_timings_seq.append(run_torch(x, residual, drop_path, dropout, norm))
        flash_timings_seq.append(run_flash(x, residual, drop_path, dropout, norm))
        tk_timings_seq.append(run_tk(x, residual, drop_path, dropout, norm))
    
    plot_timings(seq_lengths, torch_timings_seq, flash_timings_seq, tk_timings_seq,
                 'Sequence Length', 'Time (s)', 'Performance vs Sequence Length',
                 'performance_vs_sequence_length.png')

    # Varying dimension
    # b, n = 16, 2048
    # dimensions = [512, 1024, 2048, 4096]
    # torch_timings_dim = []
    # flash_timings_dim = []
    # tk_timings_dim = []
    
    # for d in dimensions:
    #     torch.manual_seed(0)
    #     x = torch.randn((b, n, d), device='cuda')
    #     residual = torch.randn((b, n, d), device='cuda')
    #     norm = nn.LayerNorm(d).cuda()
    #     dropout = nn.Dropout(p)
    #     drop_path = StochasticDepth(p, mode="row")
    #     torch_timings_dim.append(run_torch(x, residual, drop_path, dropout, norm))
    #     flash_timings_dim.append(run_flash(x, residual, drop_path, dropout, norm))
    #     tk_timings_dim.append(run_tk(x, residual, drop_path, dropout, norm))
    
    # plot_timings(dimensions, torch_timings_dim, flash_timings_dim, tk_timings_dim,
    #              'Dimension', 'Time (s)', 'Performance vs Dimension',
    #              'performance_vs_dimension.png')
    

    






import torch
import thunderkittens as tk
import random
from tqdm import tqdm
from flash_attn_interface import flash_attn_func
from einops import rearrange

def pytorch_test(Q, K, V, dO, causal):
    
    Q.requires_grad = True
    K.requires_grad = True
    V.requires_grad = True
    
    torch.backends.cuda.enable_flash_sdpa = False
    output = torch.nn.functional.scaled_dot_product_attention(Q, K, V, is_causal=causal)
    output.backward(dO)
    
    return output, Q.grad, K.grad, V.grad
    
    # Q_32 = Q.to(torch.float).requires_grad_()
    # K_32 = K.to(torch.float).requires_grad_()
    # V_32 = V.to(torch.float).requires_grad_()
    # dO_32 = dO.to(torch.float)
    
    # torch.backends.cuda.enable_flash_sdpa = True
    # output = torch.nn.functional.scaled_dot_product_attention(Q_32, K_32, V_32, is_causal=causal)
    # output.backward(dO_32)
    
    # return output.detach().to(torch.bfloat16), Q_32.grad.to(torch.bfloat16), K_32.grad.to(torch.bfloat16), V_32.grad.to(torch.bfloat16)

def h100_fwd_kernel_test(Q, K, V, dO, causal):
    Q.requires_grad = True
    K.requires_grad = True
    V.requires_grad = True
    
    o = torch.zeros_like(Q).contiguous()
    l_vec = torch.zeros(Q.shape[0], Q.shape[1], Q.shape[2], 1, device=Q.device, dtype=torch.float).requires_grad_()
    
    tk.mha_forward(Q, K, V, o, l_vec, causal)

    d_vec = torch.zeros(Q.shape[0], Q.shape[1], Q.shape[2], 1, device=Q.device, dtype=torch.float).requires_grad_()
    qg = torch.zeros_like(Q, dtype=torch.float).contiguous()
    kg = torch.zeros_like(K, dtype=torch.float).contiguous()
    vg = torch.zeros_like(V, dtype=torch.float).contiguous()

    tk.mha_backward(Q, K, V, o, l_vec, d_vec, dO, qg, kg, vg, causal)

    return o, qg, kg, vg

def flash_attn_test(Q, K, V, dO, causal):
    
    Q  = Q.permute(0, 2, 1, 3).contiguous().detach().clone()
    K  = K.permute(0, 2, 1, 3).contiguous().detach().clone()
    V  = V.permute(0, 2, 1, 3).contiguous().detach().clone()
    dO = dO.permute(0, 2, 1, 3).contiguous().detach().clone()
    
    Q.requires_grad = True
    K.requires_grad = True
    V.requires_grad = True
    
    # Forward pass
    output, _ = flash_attn_func(Q, K, V, causal=causal)
    
    # Backward pass
    output.backward(dO)
    
    Q_grad = Q.grad
    K_grad = K.grad
    V_grad = V.grad
    
    output = output.permute(0, 2, 1, 3).contiguous()
    
    Q_grad = Q_grad.permute(0, 2, 1, 3).contiguous()
    K_grad = K_grad.permute(0, 2, 1, 3).contiguous()
    V_grad = V_grad.permute(0, 2, 1, 3).contiguous()
    
    return output, Q_grad, K_grad, V_grad

def generate_tensor(shape, mean, std, dtype, device):
    return (torch.randn(*shape, dtype=dtype, device=device) * std + mean).contiguous()

def check_correctness(b, h, n, d, causal, mean, std, num_iterations=1000):
    print(f"Testing with b={b}, h={h}, n={n}, d={d}, causal={causal}")
    
    results = {
        'Output': {'tk_diff': 0, 'fa_diff': 0, 'sum_abs': 0, 'tk_max_diff': 0, 'fa_max_diff': 0},
        'Q grad': {'tk_diff': 0, 'fa_diff': 0, 'sum_abs': 0, 'tk_max_diff': 0, 'fa_max_diff': 0},
        'K grad': {'tk_diff': 0, 'fa_diff': 0, 'sum_abs': 0, 'tk_max_diff': 0, 'fa_max_diff': 0},
        'V grad': {'tk_diff': 0, 'fa_diff': 0, 'sum_abs': 0, 'tk_max_diff': 0, 'fa_max_diff': 0}
    }

    input_stats = {
        'Q': {'mean': 0, 'std': 0},
        'K': {'mean': 0, 'std': 0},
        'V': {'mean': 0, 'std': 0},
        'dO': {'mean': 0, 'std': 0}
    }

    for _ in tqdm(range(num_iterations), desc="Iterations"):
        seed = random.randint(0, 1000000)
        torch.manual_seed(seed)
        
        Q  = generate_tensor((b, h, n, d), mean=mean, std=std, dtype=torch.bfloat16, device='cuda')
        K  = generate_tensor((b, h, n, d), mean=mean, std=std, dtype=torch.bfloat16, device='cuda') 
        V  = generate_tensor((b, h, n, d), mean=mean, std=std, dtype=torch.bfloat16, device='cuda')
        dO = generate_tensor((b, h, n, d), mean=mean, std=std, dtype=torch.bfloat16, device='cuda')

        # Update input statistics
        for tensor_name, tensor in zip(['Q', 'K', 'V', 'dO'], [Q, K, V, dO]):
            input_stats[tensor_name]['mean'] += tensor.mean().item()
            input_stats[tensor_name]['std'] += tensor.std().item()

        pt_o, pt_qg, pt_kg, pt_vg = pytorch_test(Q, K, V, dO, causal)
        tk_o, tk_qg, tk_kg, tk_vg = h100_fwd_kernel_test(Q, K, V, dO, causal)
        fa_o, fa_qg, fa_kg, fa_vg = flash_attn_test(Q, K, V, dO, causal)

        tensors = [
            ('Output', pt_o, tk_o, fa_o),
            ('Q grad', pt_qg, tk_qg, fa_qg),
            ('K grad', pt_kg, tk_kg, fa_kg),
            ('V grad', pt_vg, tk_vg, fa_vg)
        ]

        for name, pt, tk, fa in tensors:
            diff_tk = pt - tk
            diff_fa = pt - fa
            abs_diff_tk = torch.abs(diff_tk)
            abs_diff_fa = torch.abs(diff_fa)
            results[name]['tk_diff'] += torch.sum(abs_diff_tk).item()
            results[name]['fa_diff'] += torch.sum(abs_diff_fa).item()
            results[name]['sum_abs'] += torch.sum(torch.abs(pt)).item()
            results[name]['tk_max_diff'] = max(results[name]['tk_max_diff'], torch.max(abs_diff_tk).item())
            results[name]['fa_max_diff'] = max(results[name]['fa_max_diff'], torch.max(abs_diff_fa).item())

        # Clear CUDA cache to free up memory
        torch.cuda.empty_cache()

    # Calculate average input statistics
    for tensor_name in input_stats:
        input_stats[tensor_name]['mean'] /= num_iterations
        input_stats[tensor_name]['std'] /= num_iterations

    print("\nResults:")
    print("=" * 80)
    print(f"Input tensor statistics (averaged over {num_iterations} iterations):")
    for tensor_name, stats in input_stats.items():
        print(f"  {tensor_name}: mean = {stats['mean']:.4e}, std = {stats['std']:.4e}")
    print("-" * 80)

    for key, data in results.items():
        tol = (data['sum_abs'] / (b * h * n * d * num_iterations)) * 0.01
        tk_avg_diff = data['tk_diff'] / (b * h * n * d * num_iterations)
        fa_avg_diff = data['fa_diff'] / (b * h * n * d * num_iterations)
        
        print(f"{key}:")
        print(f"  Max acceptable avg magnitude of diff: {tol:.2e}")
        print(f"  (A) FA 3 vs PyTorch==FA2:")
        print(f"      Avg magnitude of diff: {fa_avg_diff:.2e}")
        print(f"      Max magnitude of diff: {data['fa_max_diff']:.2e}")
        print(f"      Pass: {fa_avg_diff < tol}")
        print(f"  (B) Thunderkittens vs PyTorch==FA2:")
        print(f"      Avg magnitude of diff: {tk_avg_diff:.2e}")
        print(f"      Max magnitude of diff: {data['tk_max_diff']:.2e}")
        print(f"      Pass: {tk_avg_diff < tol}")
        print("-" * 80)

        assert fa_avg_diff < tol, f"Flash Attention 3: Average difference ({fa_avg_diff}) exceeds tolerance ({tol}) for {key}"
        assert tk_avg_diff < tol, f"Thunderkittens: Average difference ({tk_avg_diff}) exceeds tolerance ({tol}) for {key}"

print("Deep Correctness Tests:")
configurations = [
    (4, 16, 768,  128, False),
    (4, 16, 768,  128, True),
    # (4, 16, 768*2,  128, False),
    # (4, 16, 768*4,  128, False),
    # (4, 16, 768*8,  128, False),
    # (4, 16, 768*16, 128, False),
    # (4, 16, 768,    128, True),
    # (4, 16, 768*2,  128, True),
    # (4, 16, 768*4,  128, True),
    # (4, 16, 768*8,  128, True),
    # (4, 16, 768*16, 128, True),
    # (4, 32, 768,    64,  False),
    # (4, 32, 768*2,  64,  False),
    # (4, 32, 768*4,  64,  False),
    # (4, 32, 768*8,  64,  False),
    # (4, 32, 768*16, 64,  False),
    # (4, 32, 768,    64,  True),
    # (4, 32, 768*2,  64,  True),
    # (4, 32, 768*4,  64,  True),
    # (4, 32, 768*8,  64,  True),
    # (4, 32, 768*16, 64,  True),
]

for b, h, n, d, causal in configurations:
    check_correctness(b, h, n, d, causal, mean=1e-2, std=1e-1)

print("All tests passed successfully!")